import json
from collections import OrderedDict
import xml.etree.ElementTree as ET
from string import Formatter

class TXMLElement(ET.Element):
    def __init__(self, tag, attrib={}, text=None, tail=None, **extra):
        super().__init__(tag, attrib, **extra)
        self.text = text
        self.tail = tail
        
    @classmethod
    def from_dict(cls, source):
        if not isinstance(source, dict):
            raise TypeError("source must be dict, not %s" % (
                source.__class__.__name__,))
        tag = source['tag']
        attrib =source['attrib']
        xml_element = cls(tag, attrib)
        xml_element.text = source['text']
        for sub_element in source:
            xml_element.append(cls.from_dict(child))
        
        return xml_element
    
    @classmethod
    def from_element(cls, source):
        if not isinstance(source, ET.Element):
            raise TypeError("source must be ElementTree.Element, not %s" % (
                source.__class__.__name__,))
        xml_element = cls(source.tag, source.attrib)
        xml_element.text = source.text
        xml_element.tail = source.tail
        for sub_element in source:
            xml_element.append(cls.from_element(sub_element))
        return xml_element    
    
    @staticmethod
    def _to_dict(source):
        if not isinstance(source, ET.Element):
            raise TypeError("source must be ElementTree.Element, not %s" % (
                source.__class__.__name__,))        
        element = {}
        element['tag'] = source.tag
        element['attrib'] = source.attrib
        element['text'] = source.text
        element['children'] = [TXMLElement._to_dict(child) for child in source]
        return element
    
    @staticmethod
    def _to_diagram(source, indent=' '*4, show_attrib=False, level=0):
        if not isinstance(source, ET.Element):
            raise TypeError("source must be ElementTree.Element, not %s" % (
                source.__class__.__name__,)) 
        tail = ':' if show_attrib else ''
        string = '{}- {}{}\n'.format(indent*level, source.tag, tail)
        if show_attrib:
            for attrib in source.attrib:
                string += '{}- {}\n'.format(indent*(level+1), attrib)
        for child in source:
            string += TXMLElement._to_diagram(child, indent=indent, 
                                          show_attrib=show_attrib,
                                          level=level+1)
        return string
    
    
    def add_node(self, tag, attrib={}, text=None, tail=None, path=None, **extra):
        element = TXMLElement(tag, attrib, text, tail, **extra)
        self.append(element)
        return element    
    
    def to_diagram(self, indent=' '*4, show_attrib=False):
        return self._to_diagram(self, indent=indent, show_attrib=show_attrib)
        
    def to_dict(self):
        return self._to_dict(self)
    
    def __str__(self):
        return self.to_diagram(show_attrib=True)
    
    @staticmethod
    def _format_str(source, **args):
        if not isinstance(source, ET.Element):
            raise TypeError("source must be ElementTree.Element, not %s" % (
                source.__class__.__name__,)) 
        xml_str = TXMLElement._to_str(source)
        xml_formatted_str = xml_str.format(**args)
        return TXMLElement.from_str(xml_formatted_str)
    
    @staticmethod
    def _format_str_old_old(source, **args):
        if not isinstance(source, ET.Element):
            raise TypeError("source must be ElementTree.Element, not %s" % (
                source.__class__.__name__,)) 
        xml_str = json.dumps(TXMLElement._to_dict(source))
        xml_formatted_str = xml_str.format(**args)
        return TXMLElement.from_dict(xml_formatted_str)
    
    @staticmethod
    def _format_str_old(source, **args):
        if not isinstance(source, ET.Element):
            raise TypeError("source must be ElementTree.Element, not %s" % (
                source.__class__.__name__,)) 
        source.tag = source.tag.format(**args)
        if source.text:
            source.text = source.text.format(**args)
        if source.tail:
            source.tail = source.tail.format(**args)
        source.attrib = { key.format(**args): value.format(**args) for key,value in source.attrib.items()}
        for sub_element in source:
            TXMLElement._format_str(sub_element)
        
    def format_str(self, **args):
        return self._format_str(self, **args)
        
    @staticmethod
    def _iterate_append(root, source, argument_list={}):
        if not isinstance(root, ET.Element):
            raise TypeError('root node must be ElementTree.Element, not %s' % (
                root.__class__.__name__,))
        if not isinstance(source, ET.Element):
            raise TypeError('souce node must be ElementTree.Element, not %s' % (
                source.__class__.__name__,))
        element_str = ET.tostring(source, 'unicode')
        fieldnames = [fname for _, fname, _, _ in Formatter().parse(element_str) if fname]
        argnames = argument_list.keys()
        if set(fieldnames) != set(argnames):
            raise ValueError('mismatch between field names in source node and argument names in argument list')
        
        arg_size = set(map(len, argument_list.values()))
        if len(arg_size) > 1:
            raise ValueError('argument lists must have the same size')
        arg_size = arg_size[0]
        args_list = [dict(zip(argument_list,t)) for t in zip(*argument_list.values())]
        
        for args in args_list:
            element = ET.fromstring(element_str.format(**args))
            root.append(element)
        
    def iterate_append(self, source, argument_list, path=None):
        if path is None:
            root = self
        else:
            root = self.find(path)
        self._iterate_append(root, source=source, argument_list=argument_list)
        
    @staticmethod
    def _to_str(source, encoding='utf-8'):
        xml_str = ET.tostring(source)
        if not encoding:
            return xml_str
        else:
            return xml_str.decode(encoding)
    
    def to_str(self, encoding='utf-8'):
        return self._to_str(self, encoding=encoding)
    
    @classmethod
    def from_str(cls, source, encoding='utf-8'):
        if encoding:
            element = ET.fromstring(source.encode(encoding))
        else:
            element = ET.fromstring(source)
        return cls.from_element(element)
    

class TXMLTree(ET.ElementTree):
    def __init__(self, element=None, file=None,
                 version="1.0", encoding="UTF-8", doctype=None, system=None):
        super().__init__(element, file)
        if self._root is not None:
            self._root = TXMLElement.from_element(self._root)
        self._declarations = {
            'version': version,
            'encoding': encoding,
            'doctype': doctype,
            'system': system
        }
    
    def new_root(self, tag, attrib={}, text=None, tail=None, **extra):
        element = TXMLElement(tag, attrib, text, tail, **extra)
        self._root = element

    def add_node(self, tag, attrib={}, text=None, tail=None, path=None, **extra):
        if self._root is None:
            raise RuntimeError('root node is not initialized')
        element = TXMLElement(tag, attrib, text, tail, **extra)
        if path is None:
            root = self._root
        else:
            root = self._root.find(path)
        root.append(element)
        return element
        
        
    def set_root(self, element):
        if not isinstance(source, ET.Element):
            raise TypeError('root node must be ElementTree.Element, not %s' % (
                element.__class__.__name__,))
        if not isinstance(source, TXMLElement):
            self._root = TXMLElement.from_element(source)
        else:
            self._root = element
        
 
    @staticmethod
    def indent(elem, level=0):
        i = "\n" + level*"  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                TXMLTree.indent(elem, level+1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i
                
    def write(self, file_or_filename,
              encoding=None,
              xml_declaration=None,
              default_namespace=None,
              *,
              short_empty_elements=True):
        """Write element tree to a file as XML.
        Arguments:
          *file_or_filename* -- file name or a file object opened for writing
          *encoding* -- the output encoding (default: US-ASCII)
          *xml_declaration* -- bool indicating if an XML declaration should be
                               added to the output. If None, an XML declaration
                               is added if encoding IS NOT either of:
                               US-ASCII, UTF-8, or Unicode
          *default_namespace* -- sets the default XML namespace (for "xmlns")
          *short_empty_elements* -- controls the formatting of elements
                                    that contain no content. If True (default)
                                    they are emitted as a single self-closed
                                    tag, otherwise they are emitted as a pair
                                    of start/end tags
        """
        if not encoding:
            encoding = "us-ascii"
        enc_lower = encoding.lower()
        with ET._get_writer(file_or_filename, enc_lower) as write:
            if (xml_declaration or
                    (xml_declaration is None and
                     enc_lower not in ("utf-8", "us-ascii", "unicode"))):
                declared_encoding = encoding
                if enc_lower == "unicode":
                    # Retrieve the default encoding for the xml declaration
                    import locale
                    declared_encoding = locale.getpreferredencoding()
                write("<?xml version='1.0' encoding='%s'?>\n" % (
                    declared_encoding,))
            else:
                qnames, namespaces = ET._namespaces(self._root, default_namespace)
                _serialize_xml(write, self._root, qnames, namespaces,
                          short_empty_elements=short_empty_elements)
                
    def save(self, filename, **kwargs):
        if not self._root:
            return None
        with open(filename, 'wb') as f:
            f.write('<?xml version="{version}" encoding="{encoding}" ?>\n'.format(**self._declarations).encode('utf8'))
            if (self._declarations.get('doctype', None)) and (self._declarations.get('system', None)):
                f.write('<!DOCTYPE {doctype} SYSTEM "{system}">\n'.format(**self._declarations).encode('utf8'))
            self.indent(self._root)
            self.write(f, 'utf-8', **kwargs)
            
    def to_dict(self):
        if self._root is None:
            return {}
        return self._root.to_dict()     
    
    def to_diagram(self, indent=' '*4, show_attrib=False):
        if self._root is None:
            return ''
        return self._root.to_diagram(indent=indent, show_attrib=show_attrib)
            
    def __str__(self):
        if self._root is None:
            return ''
        return self._root.to_diagram()    
    
    @staticmethod
    def _to_str(source, encoding='utf-8'):
        return TXMLElement._to_str(source._root, encoding=encoding)
    
    def to_str(self, encoding='utf-8'):
        return self._to_str(self, encoding=encoding)
    
    def format_str(self, **kwargs):
        cls = TXMLTree()
        cls._declarations = self._declarations
        cls._root = self._root.format_str(**kwargs)
        return cls
    
    
def _serialize_xml(write, elem, qnames, namespaces,
                   short_empty_elements, **kwargs):
    tag = elem.tag
    text = elem.text
    if tag is ET.Comment:
        write("<!--%s-->" % text)
    elif tag is ET.ProcessingInstruction:
        write("<?%s?>" % text)
    else:
        tag = qnames[tag]
        if tag is None:
            if text:
                write(_escape_cdata(text))
            for e in elem:
                _serialize_xml(write, e, qnames, None,
                               short_empty_elements=short_empty_elements)
        else:
            write("<" + tag)
            items = list(elem.items())
            if items or namespaces:
                if namespaces:
                    for v, k in sorted(namespaces.items(),
                                       key=lambda x: x[1]):  # sort on prefix
                        if k:
                            k = ":" + k
                        write(" xmlns%s=\"%s\"" % (
                            k,
                            _escape_attrib(v)
                            ))
                for k, v in items:
                    if isinstance(k, ET.QName):
                        k = k.text
                    if isinstance(v, ET.QName):
                        v = qnames[v.text]
                    else:
                        v = ET._escape_attrib(v)
                    write(" %s=\"%s\"" % (qnames[k], v))
            if text or len(elem) or not short_empty_elements:
                write(">")
                if text:
                    write(ET._escape_cdata(text))
                for e in elem:
                    _serialize_xml(write, e, qnames, None,
                                   short_empty_elements=short_empty_elements)
                write("</" + tag + ">")
            else:
                write(" />")
    if elem.tail:
        write(ET._escape_cdata(elem.tail))


import os
import sys
import time
import json
import datetime
import functools
from concurrent.futures  import ProcessPoolExecutor

def timely_info(green_text, normal_text, adjust=40):
    print(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), '\033[92m[INFO]\033[0m', '\033[92m{}\033[0m'.format(green_text).rjust(40, ' '), normal_text)

def get_cpu_count():
    return os.cpu_count()

def parallel_run(func, *iterables, max_workers):

    with ProcessPoolExecutor(max_workers) as executor:
        result = executor.map(func, *iterables)

    return [i for i in result]

def execute_multi_tasks(func, *iterables, parallel):
    if parallel == 0:
        result = []
        for args in zip(*iterables):
            result.append(func(*args))
        return result
    else:
        if parallel == -1:
            max_workers = get_cpu_count()
        else:
            max_workers = parallel
        return parallel_run(func, *iterables, max_workers=max_workers)


def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        print('INFO: All jobs have finished. Total time taken: {:.3f} s'.format(run_time))
        return value
    return wrapper_timer

def stdout_print(msg):
    sys.__stdout__.write(msg + '\n')
    
def redirect_stdout(logfile_path):
    import ROOT
    sys.stdout = open(logfile_path, 'w')
    ROOT.gSystem.RedirectOutput(logfile_path)

def restore_stdout():
    import ROOT
    if sys.stdout != sys.__stdout__:
        sys.stdout.close()
    sys.stdout = sys.__stdout__
    ROOT.gROOT.ProcessLine('gSystem->RedirectOutput(0);')

def redirect_stdout_test(func):
    """Redirect stdout to a log file"""
    import ROOT
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        logfile_path = kwargs.get('logfile_path', None)
        if logfile_path is not None:
            sys.stdout = open(logfile_path, 'w')
            ROOT.gSystem.RedirectOutput(logfile_path)
            value = func(*args, **kwargs)
            sys.stdout.close()
            sys.stdout = sys.__stdout__
            ROOT.gROOT.ProcessLine('gSystem->RedirectOutput(0);')
            return value
        else:
            return func(*args, **kwargs)
    return wrapper_timer

def json_load(fp, *args, **kwargs):
    try:
        data = json.load(fp, *args, **kwargs)
    except:
        raise RuntimeError(f"broken json input: {fp}")
    return data


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
from matplotlib.patches import Rectangle
from matplotlib.lines import Line2D
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter,
                               AutoMinorLocator)

COLOR_PALLETE = ['#000000', '#F2385A', '#4AD9D9', '#FDC536', '#125125', '#E88EED']


PLOT_STYLES = {
    'distributions': {
        'marker': 'o',
        'markersize': 1,
        'linewidth': 1
    }
}

def suggest_markersize(nbins:int):
    if nbins <= 20:
        return 10
    elif (nbins > 20) and (nbins <= 200):
        return nbins/20
    return 1

def format_axis_ticks(ax, x_axis=True, y_axis=True, major_length=16, minor_length=8,
                      width=1, labelsize=None, x_axis_styles=None, y_axis_styles=None):
    if x_axis:
        if ax.get_xaxis().get_scale() != 'log':
            ax.xaxis.set_minor_locator(AutoMinorLocator())
        styles = {"labelsize":labelsize}
        if x_axis_styles is not None:
            styles.update(x_axis_styles)
        ax.tick_params(axis="x", which="major", direction='in', length=major_length,
                       width=width, **styles)
        ax.tick_params(axis="x", which="minor", direction='in', length=minor_length,
                       width=width, **styles)
    if y_axis:
        if ax.get_yaxis().get_scale() != 'log':
            ax.yaxis.set_minor_locator(AutoMinorLocator())    
        styles = {"labelsize":labelsize}
        if y_axis_styles is not None:
            styles.update(y_axis_styles)
        ax.tick_params(axis="y", which="major", direction='in', length=major_length,
                       width=width, **styles)
        ax.tick_params(axis="y", which="minor", direction='in', length=minor_length,
                       width=width, **styles)
    for axis in ['top','bottom','left','right']:
        ax.spines[axis].set_linewidth(width)

def centralize_axis(ax, which='x'):
    if which == 'x':
        limits = ax.get_xlim()
        xmax = np.max(np.abs(limits))
        ax.set_xlim(-xmax, xmax)
    elif which == 'y':
        limits = ax.get_ylim()
        ymax = np.max(np.abs(limits))
        ax.set_ylim(-ymax, ymax)
        
def parse_transform(obj:str):
    if not obj:
        transform = None
    elif obj == 'figure':
        transform = plt.gcf().transFigure
    elif obj == 'axis':
        transform = plt.gca().transAxes
    elif obj == 'data':
        transform = plt.gca().transData
    return transform

def get_box_dimension(box):
    axis = plt.gca()
    plt.gcf().canvas.draw()
    bb = box.get_window_extent()
    points  = bb.transformed(axis.transAxes.inverted()).get_points().transpose()
    xmin = np.min(points[0])
    xmax = np.max(points[0])
    ymin = np.min(points[1])
    ymax = np.max(points[1])
    return xmin, xmax, ymin, ymax

def draw_sigma_bands(axis, ymax, height=1.0):
    # +- 2 sigma band
    axis.add_patch(Rectangle((-2, -height/2), 2*2, ymax + height/2, fill=True, color='yellow'))
    # +- 1 sigma band
    axis.add_patch(Rectangle((-1, -height/2), 1*2, ymax + height/2, fill=True, color='lime'))
    
def draw_sigma_lines(axis, ymax, height=1.0, **styles):
    y = [-height/2, ymax*height - height/2]
    axis.add_line(Line2D([-1, -1], y, **styles))
    axis.add_line(Line2D([+1, +1], y, **styles))
    axis.add_line(Line2D([0, 0], y, **styles)) 
    
def draw_hatches(axis, ymax, height=1.0, **styles):
    x_min    = axis.get_xlim()[0]
    x_max    = axis.get_xlim()[1]
    x_range  = x_max - x_min
    y_values = np.arange(0, height*ymax, 2*height) - height/2
    for y in y_values:
        axis.add_patch(Rectangle((x_min, y), x_range, 1, **styles, zorder=-1))

def draw_text(axis, x, y, s, **styles):
    text = axis.text(x, y, s, **styles)
    xmin, xmax, ymin, ymax = get_box_dimension(text)
    return xmin, xmax, ymin, ymax

def draw_ATLAS_label(axis, x=0.05, y=0.95, fontsize=25, extra='Internal', 
                     transform_x='axis', transform_y='axis', 
                     vertical_align='top', horizontal_align='left'):
    current_axis = plt.gca()
    plt.sca(axis)
    if vertical_align not in ['top', 'bottom']:
        raise ValueError('only "top" or "bottom" vertical alignment is allowed')
    if horizontal_align not in ['left', 'right']:
        raise ValueError('only "left" or "right" horizontal alignment is allowed') 
    transform = transforms.blended_transform_factory(parse_transform(transform_x), 
                                                     parse_transform(transform_y))
    text_atlas = axis.text(x, y, 'ATLAS', fontsize=fontsize, transform=transform,
                           horizontalalignment=horizontal_align,
                           verticalalignment=vertical_align,
                           fontproperties={"weight":"bold", "style":"italic"})
    xmin, xmax, ymin, ymax = get_box_dimension(text_atlas)
    text_width = xmax - xmin
    dx = text_width/15
    text_extra = axis.text(xmax + dx, ymin, extra, fontsize=fontsize, transform=axis.transAxes,
                           horizontalalignment='left', verticalalignment='bottom')
    plt.sca(current_axis)
    _, xmax, _, ymax = get_box_dimension(text_atlas)
    return xmin, xmax, ymin, ymax


def add_collective_data(ax, data, plot_styles=None):
    if plot_styles is None:
        plot_styles = None
    color_iter = iter(COLOR_PALLETE)
    for label in data:
        styles = plot_styles.copy()
        x = data[label]['x']
        y = data[label]['y']
        nbins = len(x)
        if 'markersize' not in styles:
            styles['markersize'] = suggest_markersize(nbins)        
        if 'color' not in styles:
            styles['color'] = next(color_iter)
        styles['label'] = label
        ax.plot(x, y, **styles)

def plot_distributions(data, xlabel='', ylabel='', 
                       figsize=(14, 8), plot_styles=None, atlas_label=False):
    plt.clf()
    base_size = figsize[0]
    fig = plt.figure(figsize=figsize)
    ax = plt.gca()
    styles = PLOT_STYLES['distributions'].copy()
    if plot_styles is not None:
        styles.update(plot_styles)
    add_collective_data(ax, data, styles)
    limits = ax.get_ylim()
    ax.set_ylim(limits[0], limits[1]*1.2)
    ax.set_xlabel(xlabel, fontsize=base_size*1.6,  labelpad=base_size*0.7)
    ax.set_ylabel(ylabel, fontsize=base_size*1.6,  labelpad=base_size*0.7)
    format_axis_ticks(ax, major_length=14, minor_length=7, width=1.3, labelsize=base_size*1.4)
    if atlas_label:
        draw_ATLAS_label(ax, fontsize=base_size*1.8)
    ax.legend(fontsize=base_size*1.5)
    return plt


from typing import Optional, Union, Dict
import os
import math
import json

import numpy as np

from quickstats.components import AnalysisObject
from quickstats.utils.root_utils import load_macro
from quickstats.utils import common_utils

import ROOT

from pdb import set_trace
class AsymptoticCLs(AnalysisObject):
    
    _SIGMA_LABEL_ = {
            2: '+2sigma',
            1: '+1sigma',
            0: 'Median',
           -1: '-1sigma',
           -2: '-2sigma',
        'obs': 'Observed',
        'inj': 'Injected'
    }
    
    def __init__(self, filename:str, poi_name:str=None, data_name:str='combData', 
                 CL:float=0.95, precision:float=0.005, do_tilde:bool=True,
                 do_better_bands:bool=True, predictive_fit:bool=False,
                 better_negative_bands:bool=False, do_blind:bool=True,
                 binned_likelihood:bool=True, fix_param:str='', profile_param:str='',
                 ws_name:Optional[str]=None, mc_name:Optional[str]=None,
                 snapshot_name:Optional[str]=None, timer:bool=False,
                 minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', 
                 eps:float=1.0, strategy:int=1, print_level:int=-1, 
                 num_cpu:int=1, offset:bool=True, optimize:int=2,
                 fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1,
                 max_iters:int=-1, constrain_nuis:bool=True, batch_mode:bool=False,
                 int_bin_precision:float=-1., **kwargs):
        """
        args:
            precision: precision in mu that defines iterative cutoff
            do_tilde: bound mu at zero if true and do the \tilde{q}_{mu} asymptotics
            use_predictive_fit: experimental, extrapolate best fit nuisance parameters based on previous fit results
        """
        config = {
            'filename': filename,
            'data_name': data_name,
            'binned_likelihood' : binned_likelihood,
            'fix_param': fix_param,
            'profile_param': profile_param,
            'ws_name': ws_name,
            'mc_name': mc_name,
            'snapshot_name': snapshot_name,
            'minimizer_type': minimizer_type,
            'minimizer_algo': minimizer_algo,
            'eps': eps,
            'strategy': strategy,
            'num_cpu': num_cpu,
            'offset': offset,
            'optimize': optimize,
            'eigen': False,
            'print_level': print_level,
            'timer': timer,
            'fix_cache': fix_cache,
            'fix_multi': fix_multi,
            'max_calls': max_calls,
            'max_iters': max_iters,
            'constrain_nuis': constrain_nuis,
            'batch_mode': batch_mode,
            'int_bin_precision': int_bin_precision
        }
        
        super().__init__(**config)
        
        self.do_blind = do_blind
        self.CL = CL
        self.precision = precision
        self.do_tilde = do_tilde
        self.use_predictive_fit = predictive_fit
        self.do_better_bands = do_better_bands
        self.better_negative_bands = better_negative_bands
        self.direction = 1
        self.global_status = 0
        
        self.set_poi(poi_name)
        # define simplified names
        self.ws = self.model.workspace
        self.globs = self.model.global_observables
        self.pdf = self.model.pdf
        self.data = self.model.data
        self.nuis = self.model.nuisance_parameters
        if not self.nuis:
            self.nuis = ROOT.RooArgSet()
        # define names used for various objects
        names = self.model._DEFAULT_NAMES_
        self.nom_glob_name = names['nominal_globs']
        self.nom_nuis_name = names['nominal_nuis']
        self.nll_snapshot_name = names['nll_snapshot']
        
        self.load_extension()
        self.LimitTool = ROOT.AsymptoticCLsTool(1 - self.CL, precision, do_tilde)

    @staticmethod
    def load_extension():
        try:
            if not hasattr(ROOT, 'AsymptoticCLsTool'):
                result = load_macro('AsymptoticCLsTool')
                if hasattr(ROOT, 'AsymptoticCLsTool'):
                    print('INFO: Loaded extension module "AsymptoticCLsTool"')
        except Exception as e:
            print(e)
            
    def set_poi(self, poi_name:Optional[str]=None):
        if poi_name is None:
            self.poi = self.model.pois.first()
            print('INFO: POI name not given. Set to first poi "{}" by default.'.format(self.poi.GetName()))
        else:
            self.poi = self.model.get_poi(poi_name)
            print('INFO: POI set to "{}"'.format(poi_name))
        
    def set_poi_value(self, val:float):
        if math.isnan(val):
            raise ValueError("cannot set poi to nan")
        if ( val > 0 ) and (self.poi.getMax() < val):
            self.poi.setMax(2*val)
        if ( val < 0 ) and (self.poi.getMin() > val):
            self.poi.setMin(2*val)
        self.poi.setVal(val)
        
    def save_snapshot(self, name:str, target:str="nuis"):
        if target == "nuis":
            self.ws.saveSnapshot(name, self.nuis)
        elif target == "globs":
            self.ws.saveSnapshot(name, self.globs)
            
    def load_snapshot(self, name:str):
        snapshot = self.ws.getSnapshot(name)
        if not snapshot:
            raise RuntimeError('snapshot "{}" does not exist'.format(name))
        #print('INFO: Loaded snapshot "{}"'.format(name))
        self.ws.loadSnapshot(name)
            
    def save_nll_snapshot(self, nll, mu):
        name = self.nll_snapshot_name.format(nll_name=nll.GetName(), mu=mu)
        self.ws.saveSnapshot(name, self.nuis)
        
    def load_nll_snapshot(self, nll, mu):
        name = self.nll_snapshot_name.format(nll_name=nll.GetName(), mu=mu)
        self.load_snapshot(name) 
    
    def load_conditional_snapshot(self, mu:float, target:str="nuis"):
        name = self.get_conditional_snapsot_name(mu, target)
        self.load_snapshot(name)
    
    def get_conditional_snapsot_name(self, mu:float, target:str="nuis"):
        if target == "nuis":
            name = self.model._DEFAULT_NAMES_['conditional_nuis'].format(mu=mu)
        elif target == "globs":
            name = self.model._DEFAULT_NAMES_['conditional_globs'].format(mu=mu)
        else:
            raise ValueError('unknown conditional snapshot target "{}"'.format(target))
        return name

    def get_nll_val(self, nll:"ROOT.RooNLLVar"):
        snapshot_name = self.nll_maps[nll].get('snapshot', None)
        if snapshot_name:
            self.load_snapshot(snapshot_name)
        else:
            raise RuntimeError('Failed to load snapshot for nll "{}"'.format(nll.GetName()))
        self.global_status += self.minimizer.minimize(nll)
        self.minimizer_calls += 1
        nll_val = nll.getVal()
        self.load_snapshot(self.nom_glob_name)
        return nll_val
    
    def create_nll(self, dataset:"ROOT.RooDataSet"):
        return self.pdf.createNLL(dataset, *self.minimizer.nll_command_list)
        
    def get_qmu(self, nll:"ROOT.RooNLLVar", mu:float):
        nll_muhat = self.nll_maps[nll].get('nll_mu_hat', None)
        if nll_muhat is None:
            raise RuntimeError('Failed to get nominal nll for "{}"'.format(nll.GetName()))
        is_const = self.poi.isConstant()
        self.poi.setConstant(1)
        self.set_poi_value(mu)
        nll_value = self.get_nll_val(nll)
        self.poi.setConstant(is_const)
        qmu = 2*(nll_value - nll_muhat)
        if qmu < 0 :
            raise RuntimeError("qmu is negative which means the globally minimized "
                               "NLL is larger then the NLL of interest")
        return qmu

    def do_predictive_fit(self, nll:"ROOT.RooNLLVar", mu1:float, mu2:float, mu:float):
        if (abs(mu2-mu) < self.direction*mu*self.precision*4):
            self.load_nll_snapshot(nll, mu2)
            return None
        if not self.nuis:
            return None
        # extrapolate to mu using mu1 and mu2 assuming nuis scale linear in mu
        n_nuis = self.nuis.getSize()
        self.load_nll_snapshot(nll, mu1)
        theta_mu1 = np.array([v.getVal() for v in self.nuis])
        self.load_nll_snapshot(nll, mu2)
        theta_mu2 = np.array([v.getVal() for v in self.nuis])
        m = (theta_mu2 - theta_mu1)/(mu2-mu1)
        theta_extrap = m*mu + theta_mu2 - m*mu2
        for i, param in enumerate(self.nuis):
            param.setVal(theta_extrap[i])
            
    def get_limit(self, nll:"ROOT.RooNLLVar", initial_guess:float, summary_label:Optional[int]=None):
        self.global_status = 0
        nll_name = nll.GetName()
        print("----------------------------------")
        print("Getting limit for nll: ", nll_name)
        self.poi.setConstant(0)
        asimov_0_nll = self.summary[0]['nll']
        if nll == asimov_0_nll:
            self.set_poi_value(0)
            self.poi.setConstant(1)
        # case for observed limit
        if (nll not in self.nll_maps) or ('nll_mu_hat' not in self.nll_maps[nll]):
            nll_val = self.get_nll_val(nll)
            mu_hat = self.poi.getVal()           
            self.save_nll_snapshot(nll, mu_hat)
            if (mu_hat < 0) and (self.do_tilde):
                self.set_poi_value(0)
                self.poi.setConstant(1)
                nll_val = self.get_nll_val(nll)
            self.nll_maps[nll]['mu_hat'] = mu_hat
            self.nll_maps[nll]['nll_mu_hat'] = nll_val
        else:
            mu_hat = self.nll_maps[nll]['mu_hat']
        
        if (mu_hat < 0.1) or (initial_guess != 0):
            self.set_poi_value(initial_guess)
        
        mu = self.poi.getVal()
        qmu = self.get_qmu(asimov_0_nll, mu)
        sigma_guess = self.LimitTool.getSigma(mu, 0, qmu)
        sigma_b = sigma_guess
        mu_guess = self.LimitTool.findCrossing(sigma_guess, sigma_b, mu_hat)
        pmu = self.LimitTool.calcPmu(qmu, sigma_b, mu_guess)
        pb = self.LimitTool.calcPb(qmu, sigma_b, mu_guess)
        CLs = self.LimitTool.calcCLs(qmu, sigma_b, mu_guess)
        qmu95 = self.LimitTool.getQmu95(sigma_b, mu_guess)
        self.set_poi_value(mu_guess)
        print("Initial guess:  ", mu_guess)
        print("Sigma(obs):     ", sigma_guess)
        print("Sigma(mu,0):    ", sigma_b)
        print("muhat:          ", mu_hat)
        print("qmu95:          ", qmu95)
        print("qmu:            ", qmu)
        print("pmu:            ", pmu)
        print("1-pb:           ", pb)
        print("CLs:            ", CLs)

        n_damping = 1
        guess_to_corr = {}
        damping_factor = 1.0
        n_iter = 0
        mu_pre = mu_hat
        mu_pre2 = mu_hat
        
        while (abs(mu_pre-mu_guess) > self.precision*mu_guess*self.direction):
            print('----------------------------------')
            print('Starting iteration {} of {}'.format(n_iter, nll_name))
            # do this to avoid comparing multiple minima in the conditional and unconditional fits           
            if (n_iter == 0):
                self.load_nll_snapshot(nll, mu_hat)
            elif self.use_predictive_fit:
                self.do_predictive_fit(nll, mu_pre2, mu_pre, mu_guess)
            else:
                self.load_nll_snapshot(asimov_0_nll, mu_pre)
            qmu = self.get_qmu(nll, mu_guess)
            sigma_guess = self.LimitTool.getSigma(mu_guess, mu_hat, qmu)
            self.save_nll_snapshot(nll, mu_guess)
            
            if (nll != asimov_0_nll):
                if (n_iter == 0):
                    self.load_nll_snapshot(asimov_0_nll, self.nll_maps[asimov_0_nll]['mu_hat'])
                elif self.use_predictive_fit:
                    if (n_iter == 1):
                        self.do_predictive_fit(nll, self.nll_maps[nll]['mu_hat'], mu_pre, mu_guess)
                    else:
                        self.do_predictive_fit(nll, mu_pre2, mu_pre, mu_guess)
                else:
                    self.load_nll_snapshot(asimov_0_nll, mu_pre)
                qmuA = self.get_qmu(asimov_0_nll, mu_guess)
                sigma_b = self.LimitTool.getSigma(mu_guess, 0, qmuA)
                self.save_nll_snapshot(asimov_0_nll, mu_guess)
            else:
                sigma_b = sigma_guess
                qmuA = qmu
            corr = damping_factor*(mu_guess - self.LimitTool.findCrossing(sigma_guess, sigma_b, mu_hat))
            for first, second in guess_to_corr.items():
                if (abs(first - (mu_guess - corr)) < self.direction*mu_guess*0.02) and \
                (abs(corr) > self.direction*mu_guess*self.precision):
                    damping_factor *= 0.8
                    print('Changing damping factor to {}, n_damping = {}'.format(damping_factor, n_damping))
                    n_damping += 1
                    if n_damping > 10:
                        n_damping = 1
                        damping_factor = 1.0
                    corr *= damping_factor
                    break
            # subtract off the difference in the new and damped correction
            guess_to_corr[mu_guess] = corr
            mu_pre2 = mu_pre
            mu_pre = mu_guess
            mu_guess -= corr

            pmu = self.LimitTool.calcPmu(qmu, sigma_b, mu_pre)
            pb = self.LimitTool.calcPb(qmu, sigma_b, mu_pre)
            CLs = self.LimitTool.calcCLs(qmu, sigma_b, mu_pre)
            qmu95 = self.LimitTool.getQmu95(sigma_b, mu_pre)
            
            print("NLL:            ", nll_name)
            print("Previous guess: ", mu_pre)
            print("Sigma(obs):     ", sigma_guess)
            print("Sigma(mu,0):    ", sigma_b)
            print("muhat:          ", mu_hat)
            print("pmu:            ", pmu)
            print("1-pb:           ", pb)
            print("CLs:            ", CLs)
            print("qmu95:          ", qmu95)
            print("qmu:            ", qmu)
            print("qmuA0:          ", qmuA)
            print("Precision:      ", self.direction*mu_guess*self.precision)
            print("Correction:     ", -corr)
            print("New guess:      ", mu_guess)
            
            n_iter += 1
            if (n_iter > 25):
                raise RuntimeError('infinite loop detected in get_limit()')
        
        if summary_label is not None:
            summary = {
                'NLL_name': nll_name,
                'limit': mu_guess,
                'sigma_obs': sigma_guess,
                'sigma_b': sigma_b,
                'muhat': mu_hat,
                'pmu': pmu,
                '1-pb': pb,
                'CLs': CLs,
                'qmu95': qmu95,
                'qmu': qmu,
                'qmuA0': qmuA,
                'precision': self.direction*mu_guess*self.precision,
                'correction': -corr,
                'status': self.global_status
            }
            if summary_label not in self.summary:
                self.summary[summary_label] = summary
            else:
                self.summary[summary_label].update(summary)
        print('Found limit for nll {}: {}'.format(nll_name, mu_guess))
        print('Finished in {} itreations'.format(n_iter))
        return mu_guess
    
    @staticmethod
    def get_approx_limits(median_limit:float, target_cls:float=0.05):
        limits = {}
        sigma = median_limit/math.sqrt(3.84)
        limits[0]  = median_limit
        limits[2]  = sigma*(ROOT.Math.gaussian_quantile(1 - target_cls*ROOT.Math.gaussian_cdf( 2), 1) + 2)
        limits[1]  = sigma*(ROOT.Math.gaussian_quantile(1 - target_cls*ROOT.Math.gaussian_cdf( 1), 1) + 1)
        limits[-1] = sigma*(ROOT.Math.gaussian_quantile(1 - target_cls*ROOT.Math.gaussian_cdf(-1), 1) - 1)
        limits[-2] = sigma*(ROOT.Math.gaussian_quantile(1 - target_cls*ROOT.Math.gaussian_cdf(-2), 1) - 2)
        return limits
    
    def reset(self):
        self.global_status = 0
        self.minimizer_calls = 0
        self.approx_limits = {}
        self.limits = {}
        # mapping from RooNLLVar to other objects
        self.nll_maps = {} 
        # mapping from RooDataSet to RooNLLVar (and other objects)
        self.dataset_maps = {}
        # summary for various sigma values
        self.summary = {}
        
    def fill_mappings(self, nll:"ROOT.RooNLLVar", dataset:"ROOT.RooDataSet",
                      mu_hat:float, nll_mu_hat:float):
        snapshot_name = self.get_conditional_snapsot_name(mu_hat, target="globs")
        self.nll_maps[nll] = {}
        self.nll_maps[nll]['dataset'] = dataset
        self.nll_maps[nll]['snapshot'] = snapshot_name
        self.nll_maps[nll]['mu_hat'] = mu_hat
        self.nll_maps[nll]['nll_mu_hat'] = nll_mu_hat
        self.dataset_maps[dataset] = nll
        
    def print_failures(self):
        has_failtures = any(self.summary[sigma].get('status', 0) > 0 for sigma in self.summary)
        if has_failtures:
            print('-----------------------------------------------')
            print('Unresolved fit failures detected')
            for sigma in self.summary:
                label = self._SIGMA_LABEL_.get(sigma, sigma)
                print("{}:".format(label).ljust(10), self.summary[sigma].get('status', None))
            print('-----------------------------------------------')

    def print_summary(self):
        if (not self.limits) or (not self.approx_limits):
            print("No limits evaluated")
            return None
        print('')
        self.print_failures()
        if self.do_better_bands:
            print('Guess for bands')
        print('+2sigma:  ', self.approx_limits[2])
        print('+1sigma:  ', self.approx_limits[1])
        print('-1sigma:  ', self.approx_limits[-1])
        print('-2sigma:  ', self.approx_limits[-2])
        if self.do_better_bands:
            print('\nCorrect bands')
            print('+2sigma:  ', self.limits[2])
            print('+1sigma:  ', self.limits[1])
            print('-1sigma:  ', self.limits[-1])
            print('-2sigma:  ', self.limits[-2])
        print('Injected: ', self.limits['inj'])
        print('Median:   ', self.limits[0])
        print('Observed: ', self.limits['obs'])
        print('')
        
    def save(self, filename:str='limits.json', summary=True):
        json.dump(self.limits, open(filename, 'w'), indent=2)
        if summary:
            data = {k:{kk:vv for kk,vv in self.summary[k].items() if kk not in ['nll', 'dataset']} \
                    for k,v in self.summary.items()}
            summary_filename = os.path.splitext(filename)[0]+'_summary'+os.path.splitext(filename)[1]
            json.dump(data, open(summary_filename, 'w'), indent=2)
        
    def evaluate_limit_band(self, n:int, med_limit:float):
        init_target_cls = self.LimitTool.getTargetCLs()
        self.LimitTool.setTargetCLs(2*(1 - ROOT.Math.gaussian_cdf(abs(n))))
        if n < 0:
            self.direction = -1
            self.LimitTool.setDirection(-1)
        asimov_0_nll = self.summary[0]['nll']
        mu_guess = n*med_limit/math.sqrt(3.84)
        n_times_sigma = self.get_limit(asimov_0_nll, mu_guess)
        status = self.global_status
        sigma = n_times_sigma/n
        print('Found N * sigma = {} * {}'.format(n, sigma))
        snapshot_0_name = self.nll_maps[asimov_0_nll]['snapshot']
        self.load_snapshot(snapshot_0_name)
        asimov_data_n = self.model.generate_asimov(self.poi.GetName(), poi_val=n_times_sigma,
                                                   do_conditional=True)
        asimov_n_nll = self.create_nll(asimov_data_n)
        self.load_conditional_snapshot(n_times_sigma, target="globs")
        self.load_conditional_snapshot(n_times_sigma, target="nuis")
        self.set_poi_value(n_times_sigma)
        nll_value = asimov_n_nll.getVal()
        self.save_nll_snapshot(asimov_n_nll, n_times_sigma)
        if (n < 0) and (self.do_tilde):
            self.set_poi_value(0)
            self.poi.setConstant(1)
            nll_value = self.get_nll_val(asimov_n_nll)
        self.fill_mappings(asimov_n_nll, asimov_data_n, n_times_sigma, nll_value)
        self.LimitTool.setTargetCLs(init_target_cls)
        self.direction = 1
        self.LimitTool.setDirection(1)
        mu_guess = self.LimitTool.findCrossing(sigma, sigma, n_times_sigma)
        limit = self.get_limit(asimov_n_nll, mu_guess, n)
        return limit    
    
    @common_utils.timer
    def evaluate_limits(self):
        # TODO: maybe load nominal snapshot first
        self.reset()
        obs_nll = self.create_nll(self.data)
        self.nll_maps[obs_nll] = {"dataset": self.data, "snapshot": self.nom_glob_name}
        self.dataset_maps[self.data] = obs_nll
        
        self.save_snapshot(self.nom_glob_name, target="globs")
        self.save_snapshot(self.nom_nuis_name, target="nuis")

        asimov_data_0 = self.model.generate_asimov(self.poi.GetName(), poi_val=0,
                                                   do_conditional=not self.do_blind,
                                                   poi_profile=0, dataset=self.data)
        asimov_0_nll = self.create_nll(asimov_data_0)
        self.summary[0] = {'dataset': asimov_data_0, 'nll': asimov_0_nll}
        self.set_poi_value(0)
        self.save_nll_snapshot(asimov_0_nll, 0)
        self.load_conditional_snapshot(0, target="globs")
        self.load_conditional_snapshot(0, target="nuis")
        nll_value = asimov_0_nll.getVal()
        self.fill_mappings(asimov_0_nll, asimov_data_0, 0, nll_value)
        
        self.LimitTool.setTargetCLs(1 - self.CL)
        
        med_limit = self.get_limit(asimov_0_nll, 1.0, 0)
        approx_limits = self.get_approx_limits(med_limit, self.LimitTool.getTargetCLs())
        limits = approx_limits.copy()
        
        sigma = med_limit/math.sqrt(3.84)
        self.poi.setRange(-5*sigma, 5*sigma)
        
        # compute better sigma bands
        bands_to_improve = []
        if self.do_better_bands:
            if self.better_negative_bands:
                bands_to_improve = [2, 1, -1, -2]
            else:
                bands_to_improve = [2, 1]
        
        for n in bands_to_improve:
            limits[n] = self.evaluate_limit_band(n, med_limit)
        
        # compute observed limit
        self.load_conditional_snapshot(0, target="nuis")
        obs_limit = 0 if self.do_blind else self.get_limit(obs_nll, med_limit, 'obs')
        limits['obs'] = obs_limit
        
        # injection test not supported yet
        limits['inj'] = 0
        
        self.limits = limits
        self.approx_limits = approx_limits
        
        self.print_summary()
        
        print('INFO: Finished with {} calls to minimize(nll)'.format(self.minimizer_calls))
        return limits


#from quickstats.components.cli_parser import CLIParser
#from quickstats.components.np_ranking_plotter import NPRankingPlotter
from quickstats.components.extended_minimizer import ExtendedMinimizer
from quickstats.components.extended_model import ExtendedModel
from quickstats.components.nuisance_parameter_pull import NuisanceParameterPull
from quickstats.components.roofit_plotter import RooFitPlotter
from quickstats.components.nuisance_parameter_harmonizer import NuisanceParameterHarmonizer
from quickstats.components.analysis_object import AnalysisObject
from quickstats.components.asymptotic_cls import AsymptoticCLs


import click

import fnmatch
from quickstats.components import ExtendedModel
    
@click.group(name='inspect_ws')
def inspect_ws():
    '''
        Tool for inspecting workspace attributes
    '''
    pass

def get_ws_attributes(attrib_name, input_file, ws_name, mc_name, output_file=None, patterns=None):
    attrib_func_map = {
        'pois': ExtendedModel.get_poi_names,
        'datasets': ExtendedModel.get_dataset_names,
        'nuisance_parameters': ExtendedModel.get_nuisance_parameter_names
    }
    attrib_func = attrib_func_map.get(attrib_name, None)
    if attrib_func is None:
        raise ValueError('invalid workspace attribute {}'.format(attrib_name))
    attributes = attrib_func(input_file, ws_name, mc_name)
    if output_file is None:
        print(attributes)
    else:
        with open(output_file, 'w') as output:
            output.write('\n'.join(attributes) + '\n')
    if patterns is not None:
        patterns = patterns.split(',')
        matched = sorted(list(set([attr for pattern in patterns for attr in attributes if fnmatch.fnmatch(attr, pattern)])))
        escaped = sorted(list(set(attributes) - set(matched)))
        print("INFO: Attributes that match the given pattern(s):")
        print(matched)
        print("INFO: Attributes that escaped the given pattern(s):")
        print(escaped)

@inspect_ws.command(name='nuisance_parameters')
@click.option('-i', '--input_file', required=True, help='Path to the input workspace file')
@click.option('-w', '--workspace', 'ws_name', default=None, help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, help='Name of model config. Auto-detect by default.')
@click.option('-o', '--output_file', default=None, help='Export output to text file')
@click.option('-p', '--patterns', default=None, help='Match nuisance parameters with given patterns (separated by commas).')
def get_ws_nuisance_parameters(input_file, ws_name, mc_name, output_file=None, patterns=None):
    get_ws_attributes('nuisance_parameters', input_file, ws_name, mc_name, output_file, patterns)
    
@inspect_ws.command(name='datasets')
@click.option('-i', '--input_file', required=True, help='Path to the input workspace file')
@click.option('-w', '--workspace', 'ws_name', default=None, help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, help='Name of model config. Auto-detect by default.')
@click.option('-o', '--output_file', default=None, help='Export output to text file')
@click.option('-p', '--patterns', default=None, help='Match datasets with given patterns (separated by commas).')
def get_ws_datasets(input_file, ws_name, mc_name, output_file=None, patterns=None):
    get_ws_attributes('datasets', input_file, ws_name, mc_name, output_file, patterns)
    
@inspect_ws.command(name='pois')
@click.option('-i', '--input_file', required=True, help='Path to the input workspace file')
@click.option('-w', '--workspace', 'ws_name', default=None, help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, help='Name of model config. Auto-detect by default.')
@click.option('-o', '--output_file', default=None, help='Export output to text file')
@click.option('-p', '--patterns', default=None, help='Match POIs with given patterns (separated by commas).')
def get_ws_pois(input_file, ws_name, mc_name, output_file=None, patterns=None):
    get_ws_attributes('pois', input_file, ws_name, mc_name, output_file, patterns)    


from typing import List
from copy import deepcopy
from quickstats.components.root_plots import set_attrib
        
import ROOT

class RooFitPlotter(object):
    
    _DEFAULT_BOX_OPTIONS_ = {
        'FillColor'  : 0,
        'BorderSize' : 0,
        'TextAlign'  : 12,
        'TextSize'   : 0.04,
        'FillStyle'  : 1001
    }
    
    _DEFAULT_TEXTS_ = ['ATLAS #bf{Internal}', '#sqrt{s} = 13 TeV, L=139fb^{-1}']

    def __init__(self, x, data, model, name='roofit_plot'):
        self.x = x
        self.data = data
        self.model = model
        self.name = name
        
    def setup_canvas(self):
        canvas_name = 'c_{}'.format(self.name)
        if not ROOT.gROOT.FindObject(canvas_name):
            self.canvas = ROOT.TCanvas(canvas_name, canvas_name,800,600)
        else:
            self.canvas = ROOT.gROOT.FindObject(canvas_name)

        self.canvas.Clear()
        self.canvas.Divide(1, 2)
        self.canvas.GetPad(1).SetPad(0.0, 0.25, 1, 1)
        self.canvas.GetPad(1).SetBottomMargin(0.05)
        self.canvas.GetPad(1).SetRightMargin(0.05)
        self.canvas.GetPad(1).SetTicks(1, 1)
        self.canvas.GetPad(2).SetPad(0.0, 0.0, 1, 0.25)
        self.canvas.GetPad(2).SetBottomMargin(0.32)
        self.canvas.GetPad(2).SetTopMargin(0.05)
        self.canvas.GetPad(2).SetRightMargin(0.15)
        self.canvas.GetPad(2).SetRightMargin(0.05)
        self.canvas.GetPad(2).SetTicks(1, 1)
        
        self.legend = ROOT.TLegend(0.63, 0.70, 0.85, 0.85)
        self.legend.SetBorderSize(0)
        self.label = ROOT.TLatex()
    
    def plot_text(self, texts, name=None, box_options=None, x_min=0.1,
                  x_max=0.3, y_max=0.85, dy=0.06):
        
        n_lines = len(texts)
        y_min = y_max - dy * n_lines
        
        # create text box
        box = ROOT.TPaveText(x_min, y_max, x_max, y_min,"BRNDC");
        if not box:
            return 0
        if name is None:
            box.SetName('new_textBox')
        else:
            box.SetName(name)
        
        set_attrib(box, **self._DEFAULT_BOX_OPTIONS_)
        if box_options:
            set_attrib(box, **box_options)
        
        # add labels
        for text in texts:
            box.AddText(text)
    
        self.frame.addObject(box)
    
        return box
        
    def plot_param(self, params=None, show_constants=False, labels=None,
                   name=None, box_options=None, precision=3,
                   x_min=0.6, x_max=0.99, y_max=0.85, dy=0.06):
        
        parameters = self.model.getParameters(self.data)
        if params is not None:
            parameters = [p for p in parameters if p.GetName() in params]
        if not show_constants:
            parameters = [p for p in parameters if not p.isConstant()]
        param_format = ROOT.RooFit.Format("NEU", ROOT.RooFit.AutoPrecision(precision))
        parameter_text = [p.format(param_format).Data() for p in parameters]
        
        if labels:
            parameter_text += labels
            
        if not name:
            name = '{}_paramBox'.format(self.model.GetName())
            
        return self.plot_text(parameter_text, name=name, box_options=box_options,
                              x_min=x_min, x_max=x_max, y_max=y_max, dy=dy)
        
    @staticmethod
    def get_relative_hist(frame, histname, curvename, threshold=None, epsilon=1e-7):
        hist = frame.findObject(histname)
        curve = frame.findObject(curvename)
        rel_hist = ROOT.RooHist(hist.getNominalBinWidth())
        #rel_hist.SetName('relative_error_{}_{}'.format(hist.GetName(), curve.GetName()))
        #rel_hist.SetTitle('relative error of {} and {}'.format(hist.GetTitle(), curve.GetTitle()))

        #Determine range of curve
        xstart, xstop, y = ROOT.Double(), ROOT.Double(), ROOT.Double()
        curve.GetPoint(0, xstart, y)
        curve.GetPoint(curve.GetN()-1, xstop, y)
        n_bin = hist.GetN()
        for i in range(n_bin):
            x = ROOT.Double()
            point = ROOT.Double()
            hist.GetPoint(i, x, point)
            if (x < xstart) or (x > xstop):
                continue
            yy = ROOT.Double()
            #yy = 1 - data/fit
            fit_value = curve.interpolate(x)
            if abs(point) <= epsilon:
                yy = 0
            else:
                yy = (fit_value - point)/point
            if threshold and abs(yy) > threshold:
                continue
            if abs(point) <= epsilon:
                dyl = 0
                dyh = 0
            else:
                dyl = hist.GetErrorYlow(i)/fit_value
                dyh = hist.GetErrorYhigh(i)/fit_value
            rel_hist.addBinWithError(x, yy, dyl, dyh)
        return rel_hist
    
    def create_ratio(self, x_title="m_{#gamma#gamma}[GeV]", label_scale=1.0,
                     marker_size = 0.4, threshold = 1.0):
        self.canvas.cd(2)
        self.h_ratio = self.get_relative_hist(self.frame, 'data', 'fit',
                                              threshold=threshold)
        self.h_ratio.SetMarkerSize(0.4)
        self.h_ratio.SetMarkerColor(ROOT.kRed)
        self.ratio_frame = self.x.frame(ROOT.RooFit.Title(" "))
        self.ratio_frame.addPlotable(self.h_ratio, "P")
        
        self.ratio_frame.SetXTitle(x_title)
        self.ratio_frame.GetYaxis().SetTitle("Relative Error")
        self.ratio_frame.GetYaxis().CenterTitle()
        self.ratio_frame.GetXaxis().SetTitleSize(0.12)
        self.ratio_frame.GetYaxis().SetTitleSize(0.1)
        self.ratio_frame.GetYaxis().SetTitleOffset(0.4)
        self.ratio_frame.GetXaxis().SetTitleOffset(1.3)
        self.ratio_frame.GetXaxis().SetLabelSize(0.12 * label_scale)
        self.ratio_frame.GetYaxis().SetLabelSize(0.12 * label_scale)
        self.ratio_frame.GetYaxis().SetLabelOffset(0.006)
        self.ratio_frame.GetYaxis().SetRangeUser(-1.0, 1.0)
        
        self.ratio_frame.GetXaxis().SetLabelFont(2)
        self.ratio_frame.GetXaxis().SetTitleFont(2)
        self.ratio_frame.GetXaxis().SetLabelFont(2)
        
    def format_ticks(self, tick_len=20):
        ### Equal sized ticks!!
        pad1 = self.canvas.GetPad(1)
        pad2 = self.canvas.GetPad(2)

        pad1W = pad1.GetWw() * pad1.GetAbsWNDC()
        pad1H = pad1.GetWh() * pad1.GetAbsHNDC()
        pad2W = pad2.GetWw() * pad2.GetAbsWNDC()
        pad2H = pad2.GetWh() * pad2.GetAbsHNDC()
        
        #self.ratio_frame.GetXaxis().SetTickLength(self.ratio_frame.GetXaxis().GetTickLength() * 3.0)
       
        self.frame.GetYaxis().SetNdivisions(1005)
        self.ratio_frame.GetYaxis().SetNdivisions(505)
        self.frame.GetXaxis().SetNdivisions(1010)
        self.ratio_frame.GetXaxis().SetNdivisions(1010)

        self.frame.SetTickLength(tick_len/pad1W, "Y")
        self.frame.SetTickLength(tick_len/pad1H, "X")
        self.ratio_frame.SetTickLength(tick_len/pad2W, "Y")
        self.ratio_frame.SetTickLength(tick_len/pad2H, "X")
        
    def draw_labels(self, labels):
        self.canvas.cd(1)
        self.label.SetTextSize(0.1);
        self.label.SetTextAlign(13);
        for i, label in enumerate(labels):
            self.label.DrawLatex(0.36, 0.9-0.0275*i, label)
        
    def plot(self, x_title="m_{#gamma#gamma}[GeV]",
             marker_size = 0.3, line_width=2, label_scale=1.0,
             ratio_marker_size = 0.4, labels=None):
        self.setup_canvas()
        self.canvas.cd(1)
        self.frame = self.x.frame(ROOT.RooFit.Title(" "))
        self.data.plotOn(self.frame, ROOT.RooFit.Name("data"), 
                         ROOT.RooFit.MarkerSize(marker_size),
                         ROOT.RooFit.DataError(ROOT.RooAbsData.SumW2))
        self.model.plotOn(self.frame, ROOT.RooFit.Name("fit"), 
                          ROOT.RooFit.LineWidth(line_width), 
                          ROOT.RooFit.LineColor(ROOT.kRed))
        self.data.plotOn(self.frame, ROOT.RooFit.Name("data"), 
                         ROOT.RooFit.MarkerSize(marker_size),
                         ROOT.RooFit.DataError(ROOT.RooAbsData.SumW2))
        self.legend.AddEntry(self.frame.findObject("data"), "Data", "LPE")
        self.legend.AddEntry(self.frame.findObject("fit"), "Fit", "L")
        
        self.frame.GetXaxis().SetLabelFont(2)
        self.frame.GetXaxis().SetTitleFont(2)
        self.frame.GetXaxis().SetLabelFont(2)
        self.frame.GetXaxis().SetTitle(" ")
        self.create_ratio(marker_size=ratio_marker_size)
        self.format_ticks()
        
        self.canvas.cd(1)
        
        self.plot_param(y_max=0.67, name='{}_paramBox'.format(self.name))
        
        texts = deepcopy(self._DEFAULT_TEXTS_)
        if labels:
            texts += labels
        self.plot_text(texts, x_min=0.12, name='{}_textBox'.format(self.name))
        
        ##double chi2 = frame->chiSquare("fit","data",7); 
        #TString chi2Line = TString::Format("Chisquare = %f ",chi2)
        #fit->paramOn(frame,Layout(0.7,0.9,0.9), Label(chi2Line))
        
        self.frame.Draw()
        self.legend.SetTextFont(2)
        self.legend.Draw("same")
        self.canvas.cd(2)
        self.ratio_frame.Draw()
        
        return self.canvas


import os
import time
import json
from itertools import repeat
from typing import List, Optional

import numpy as np
import ROOT
from quickstats.components import ExtendedModel, ExtendedMinimizer
from quickstats.components.numerics import approx_n_digit, str_encode_value, str_decode_value
from quickstats.utils import common_utils

def load_cache(fname:str):
    if (fname is not None) and os.path.exists(fname):
        with open(fname, 'r') as file:
            data = json.load(file)
        return data
    else:
        return None

def evaluate_nll(input_file:str, poi_val:float, poi:str="", vmin:float=10., vmax:float=10., unconditional=False,
                 workspace:Optional[str]=None, model_config:Optional[str]=None, data:str='combData', snapshot:str=None, 
                 profile:str="", fix:str="", hesse:bool=False, minos:bool=True, constrain:bool=True, 
                 minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', num_cpu:int=1, binned:bool=True,
                 eps:float=1.0, strategy:int=0, fix_cache:bool=True, fix_multi:bool=True, mpsplit:int=3,
                 verbose:int=0, max_calls:int=-1, max_iters:int=-1, optimize:int=2, offset:bool=False,
                 print_level:int=1, outname:str=None, cache:bool=False, detailed_output:bool=False):
    if cache:
        cached_result = load_cache(outname)
        if cached_result is not None:
            poi_name, nll = cached_result['poi'], cached_result['nll']
            print('INFO: Found NLL cache from "{}"'.format(outname))
            if unconditional:
                print('INFO: Cached unconditional NLL for POI "{}": {}'.format(poi_name, nll))
            else:
                print('INFO: Cached NLL for POI "{}" at {:.2f}: {}'.format(poi_name, poi_val, nll))
            return nll
    start = time.time()
    model = ExtendedModel(fname=input_file, ws_name=workspace, mc_name=model_config,
                          data_name=data, binned_likelihood=binned, snapshot_name=snapshot, 
                          fix_cache=fix_cache, fix_multi=fix_multi)
    if fix:
        model.fix_parameters(fix)
    if profile:
        print('INFO: Profiling parameters')
        model.profile_parameters(profile)    
    # setup the POI
    if poi:
        poi = model.workspace.var(poi)
    else:
        poi = model.pois.first()
    poi_min, poi_max = poi_val-abs(vmin), poi_val+abs(vmax)
    poi.setRange(poi_min, poi_max)
    if unconditional:
        poi.setConstant(0)
        hesse = True
    else:
        poi.setVal(poi_val)
        poi.setConstant(1)

    minimizer = ExtendedMinimizer("minimizer", model.pdf, model.data)
    # configure minimize options
    nll_commands = [ROOT.RooFit.NumCPU(num_cpu, mpsplit), 
                    ROOT.RooFit.GlobalObservables(model.global_observables), 
                    ROOT.RooFit.Offset(offset)]

    if constrain:
        nll_commands.append(ROOT.RooFit.Constrain(model.nuisance_parameters))
    minimize_options = {
        'minimizer_type'   : minimizer_type,
        'minimizer_algo'   : minimizer_algo,
        'default_strategy' : strategy,
        'opt_const'        : optimize,
        'eps'              : eps,
        'max_calls'        : max_calls,
        'max_iters'        : max_iters,
        'hesse'            : hesse,
        'verbose'          : verbose,
        'print_level'      : print_level
    }
    if minos:
        minimize_options['minos']     = True
        minimize_options['minos_set'] = ROOT.RooArgSet(poi)
    # perform the fit
    minimizer.minimize(nll_commands=nll_commands, **minimize_options)
    nll = minimizer.fit_result.minNll()
    end = time.time()
    if unconditional:
        print('INFO: Unconditional NLL for POI "{}": {}'.format(poi.GetName(), nll))
    else:
        print('INFO: NLL for POI "{}" at {:.2f}: {}'.format(poi.GetName(), poi_val, nll))
    print('INFO: Time taken to evaluate NLL: {:.3f} s'.format(end-start))
    
    results = {
        'nll': nll,
        'poi': poi.GetName(),
        'constrain': int(constrain),
        'poi_value': poi_val,
        'poi_min': poi_min,
        'poi_max': poi_max,
        'unconditional': int(unconditional)
    }
    if unconditional:
        poi_val = poi.getVal()
        results['poi_bestfit'] = poi_val
    results['time'] = end-start
    
    # save results
    if outname is not None:
        with open(outname, 'w') as outfile:
            json.dump(results, outfile)
        print('INFO: Saved NLL result to {}'.format(outname))
        
    if detailed_output:
        return results
    return nll

def scan_nll(input_file:str, poi_min:float, poi_max:float, poi_step:float, poi:str="", cache:bool=True,
             outname:str="{poi}", outdir:str='output', vmin:float=10., vmax:float=10., 
             workspace:Optional[str]=None, model_config:Optional[str]=None, data:str='combData', snapshot:str=None, 
             profile:str="", fix:str="", hesse:bool=False, minos:bool=True, constrain:bool=True,
             minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', num_cpu:int=1, binned:bool=True,
             eps:float=1.0, strategy:int=0, fix_cache:bool=True, fix_multi:bool=True, mpsplit:int=3, 
             verbose:int=0, print_level:int=1, max_calls:int=-1, max_iters:int=-1, optimize:int=2, 
             offset:bool=False, parallel:int=8):
    start_time = time.time()
    points        = np.arange(poi_min, poi_max+poi_step, poi_step)
    unconditional = np.concatenate(([True], np.full((len(points)), False)))
    points        = np.concatenate(([0.], points))
    if not poi:
        poi = ExtendedModel.get_poi_names(input_file, workspace, model_config)[0]
        print('INFO: POI not given, default as "{}".'.format(poi))
    # try to get the number of significant digits of poi_step
    sd = approx_n_digit(poi_step)
    # save cache file as outdir/cache/outname_{str_encoded_poi_value}.json
    outnames = ["{}_{}.json".format(outname.format(poi=poi), "uncond" if uncond else str_encode_value(point, sd)) \
                for point, uncond in zip(points, unconditional)]
    cachedir = os.path.join(outdir, "cache")
    outnames = [os.path.join(cachedir, name) for name in outnames]
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    if not os.path.exists(cachedir):
        os.makedirs(cachedir)

    arguments = (repeat(input_file), points, repeat(poi), repeat(vmin), repeat(vmax),
                 unconditional, repeat(workspace),repeat(model_config), repeat(data),
                 repeat(snapshot), repeat(profile), repeat(fix), repeat(hesse), repeat(minos),
                 repeat(constrain), repeat(minimizer_type), repeat(minimizer_algo), repeat(num_cpu),
                 repeat(binned), repeat(eps), repeat(strategy), repeat(fix_cache),
                 repeat(fix_multi), repeat(mpsplit), repeat(verbose), repeat(max_calls), repeat(max_iters), 
                 repeat(optimize), repeat(offset), repeat(print_level), outnames, repeat(cache))
    results = common_utils.execute_multi_tasks(evaluate_nll, *arguments, parallel=parallel)

    uncond_nll = results[0]
    data = {"{{:.{}f}}".format(sd).format(poi_val):2*(nll-uncond_nll) \
            for poi_val, nll in zip(points[1:], results[1:])}
    final_outname = os.path.join(outdir, outname.format(poi=poi) + '.json')                  
    with open(final_outname, 'w') as outfile:
        json.dump(data, outfile, indent=3)
    end_time = time.time()
    print('INFO: All jobs have finished. Total Time taken: {:.3f} s'.format(end_time-start_time))


import os
import re
import json
import glob
import datetime
import difflib

from quickstats.utils.common_utils import timely_info

class NuisanceParameterHarmonizer(object):
    def __init__(self, ref_list=None):
        self.ref_list = ref_list

    @property
    def ref_list(self):
        return self._ref_list
    
    @ref_list.setter
    def ref_list(self, val):
        if val is None:
            self._ref_list = {}
        elif isinstance(val, list):
            self._ref_list = config
        elif isinstance(val, str):
            self._ref_list = json.load(open(val, 'r'))
        else:
            raise ValueError("invalid reference list format")
        
    @staticmethod
    def get_nuiance_parameter_names_collection(ws_paths):
        from quickstats.components import ExtendedModel
        nuis_names_collection = set()
        for ws_path in ws_paths:
            nuis_names = ExtendedModel.get_nuisance_parameter_names(ws_path)
            nuis_names_collection |= set(nuis_names)
        return sorted(list(nuis_names_collection))
    
    @staticmethod
    def filter_nuis(source, patterns=None, excludes=None, exact=False):
        if patterns is None:
            return []    
        if isinstance(patterns, str):
            patterns = [patterns]
        if exact:
            return list(set(source) & set(patterns))
        matches = []    
        for pattern in patterns:
            regex = re.compile(pattern)
            matches += [s for s in source if regex.search(s)]
        if excludes is not None:
            anti_matches = NuisanceParameterHarmonizer.filter_nuis(source, excludes)
            matches = list(set(matches) - set(anti_matches))
        return matches
    
    @staticmethod
    def get_suggested_names(source, possibilities, cutoff=0.6, replace={}):
        suggested_names = {}
        for name in source:
            _name = name
            for old, new in replace.items():
                _name = _name.replace(old, new)             
            matches = difflib.get_close_matches(_name, possibilities, n=1, cutoff=cutoff)
            if matches:
                suggested_names[name] = matches[0]
        return suggested_names
    
    @staticmethod
    def display_difference(old_name, new_name):
        sequence = difflib.SequenceMatcher(None, old_name, new_name)
        match = sequence.find_longest_match(0, len(old_name), 0, len(new_name))
        ahead = old_name[ : match.a]
        atail = old_name[match.a+match.size : ]
        bhead = new_name[ : match.b]
        btail = new_name[match.b+match.size : ]
        common1 = old_name[match.a : match.a+match.size]
        common2 = new_name[match.b : match.b+match.size]
        assert common1 == common2
        print(f'{ahead}\033[91m{common1}\033[0m{atail}'.rjust(80, ' '), '->', f'{bhead}\033[92m{common1}\033[0m{btail}')

    @staticmethod
    def display_mappings(mappings, reference, level=0):
        for k, v in mappings.items():
            description = reference[k].get('description', k)
            if reference[k].get('display', None) == False:
                continue
            print('{indent}{description}:'.format(indent='\t'*level, description=description))
            if 'source' in reference[k]:
                NuisanceParameterHarmonizer.display_mappings(v, reference[k]['source'], level=1)
            else:
                for old_name, new_name in v.items():
                    NuisanceParameterHarmonizer.display_difference(old_name, new_name)
                
    @staticmethod
    def reshape_mappings(mappings, roots=[], result=None):
        if result is None:
            result = {}
        for key, value in mappings.items():
            if isinstance(value, dict):
                NuisanceParameterHarmonizer.reshape_mappings(value, roots + [key], result)
            elif key not in result:
                result[key] = [(roots, value)]
            else:
                result[key].append((roots, value))
        return result

    @staticmethod
    def flatten_mappings(mappings):
        flattened_mappings = {}
        # check duplicates
        reshaped_mappings = NuisanceParameterHarmonizer.reshape_mappings(mappings)
        for old_name in reshaped_mappings:
            dest = reshaped_mappings[old_name]
            if len(reshaped_mappings[old_name]) > 1:
                raise RuntimeError('multiple mappings found for the NP: {} ({})'.format(old_name, 
                                   ','.join([':'.join(field[0]) for field in dest])))
            flattened_mappings[old_name] = dest[0][1]
        return flattened_mappings
    
    @staticmethod
    def remove_nuis(source, target):
        for t in target:
            source.remove(t)
        return source
        
    @staticmethod
    def match_source(suggested_names, source):
        mappings = {s:{} for s in source}
        for old_name, new_name in suggested_names.items():
            for key, value in source.items():
                if new_name in value['names']:
                    mappings[key][old_name] = new_name
        mappings = {k:v for k,v in mappings.items() if v != {}}
        return mappings
    
    @staticmethod
    def get_mappings(nuis_names, reference):
        mappings = {}
        for group in reference:
            # rename by custom mapping
            custom_map = reference[group].get('map', None)
            if custom_map is not None:
                temp = {}
                for old_name, new_name in custom_map.items():
                    if old_name in nuis_names:
                        temp[old_name] = new_name
                if temp:
                    mappings[group] = temp
                continue
                        
            selected = []
            # NPs that match a pattern with exclusion
            selected += NuisanceParameterHarmonizer.filter_nuis(nuis_names, reference[group].get('pattern', None), 
                        reference[group].get('exclude', None), exact=False)
            # NPs that match exactly
            selected += NuisanceParameterHarmonizer.filter_nuis(nuis_names, reference[group].get('match', None),
                        exact=True)
            if not selected:
                continue
            # rename by reformatting
            if ('source' not in reference[group]) and ('names' not in reference[group]):
                format_str = reference[group].get('format', '{NAME}')
                mappings[group] = {}
                for old_name in selected:
                    new_name = old_name
                    # replace strings in old NP
                    for old, new in reference[group].get('replace', {}).items():
                        new_name = new_name.replace(old, new)
                    # format new NP name
                    new_name = format_str.format(NAME=new_name)
                    mappings[group][old_name] = new_name
            # rename by matching suggestions
            else:
                cutoff = reference[group].get('cutoff', 0.6)
                if 'names' in reference[group]:
                    names = reference[group]['names']
                elif 'source' in reference[group]:
                    names = []
                    for key, value in reference[group]['source'].items():
                        names += value['names']
                replace = reference[group].get('replace', {})
                suggested_names = NuisanceParameterHarmonizer.get_suggested_names(selected, names, cutoff, replace)
                if suggested_names:
                    if 'names' in reference[group]:
                        mappings[group] = suggested_names
                    elif 'source' in reference[group]:
                        source = reference[group]['source']
                        mappings[group] = NuisanceParameterHarmonizer.match_source(suggested_names, source)
        return mappings


    def get_harmonize_map(self, nuis_names):
        mappings = self.get_mappings(nuis_names, self.ref_list)
        self.display_mappings(mappings, self.ref_list)
        flattened_mappings = self.flatten_mappings(mappings)
        unknown_nuis = list(set(nuis_names) - set(flattened_mappings))
        if len(unknown_nuis) > 0:
            print('Remained NPs:')
            for unknown in unknown_nuis:
                NuisanceParameterHarmonizer.display_difference(unknown, '?')
        flattened_mappings.update({np:None for np in unknown_nuis})
        return flattened_mappings
    
    def harmonize(self, ws_paths, outfile=None):
        if isinstance(ws_paths, str):
            ws_paths = [ws_paths]
        timely_info('Harmonizing collective workspace files', ','.join(ws_paths))
        nuis_names = NuisanceParameterHarmonizer.get_nuiance_parameter_names_collection(ws_paths)
        harmonize_map = self.get_harmonize_map(nuis_names)
        if outfile is not None:
            json.dump(harmonize_map, open(outfile, 'w'), indent=2, sort_keys=True)
        return harmonize_map
    
    def _harmonize_multi_input(self, input_paths):
        harmonize_map = {}
        if isinstance(input_paths, (str, list)):
            harmonize_map = self.harmonize(input_paths)
        elif isinstance(input_paths, dict):
            for k,v in input_paths.items():
                harmonize_map[k] = self._harmonize_multi_input(v)
        return harmonize_map
    
    @staticmethod
    def get_input_paths(input_expr, base_path='./'):
        if isinstance(input_expr, str):
            full_path = os.path.join(base_path, input_expr)
            ws_files = glob.glob(full_path)
            if not ws_files:
                print('WARNING: No matching input found in {}'.format(full_path))
            return ws_files
        elif isinstance(input_expr, list):
            input_paths = []
            for expr in input_expr:
                input_paths += NuisanceParameterHarmonizer.get_input_paths(expr, base_path)
            return input_paths
        elif isinstance(input_expr, dict):
            input_paths = {}
            for k,v in input_expr.items():
                input_paths[k] = NuisanceParameterHarmonizer.get_input_paths(v, base_path)
            return input_paths
        else:
            raise ValueError('invalid input expr')
        
    def harmonize_multi_input(self, input_config_path, base_path='./', outfile=None):
        config = json.load(open(input_config_path))
        input_paths = NuisanceParameterHarmonizer.get_input_paths(config, base_path)
        harmonize_map = self._harmonize_multi_input(input_paths)
        if outfile is not None:
            json.dump(harmonize_map, open(outfile, 'w'), indent=2, sort_keys=True)
            timely_info('Saved rename output as ', outfile)
        return harmonize_map



import os
import re
import setuptools
from setuptools.command.install import install

def get_macros_dir():
    """Return the installation directory, or None"""
    if '--user' in sys.argv:
        paths = (site.getusersitepackages(),)
    else:
        py_version = '%s.%s' % (sys.version_info[0], sys.version_info[1])
        paths = (s % (py_version) for s in (
            sys.prefix + '/lib/python%s/site-packages/',
            sys.prefix + '/local/lib/python%s/site-packages/',
            sys.prefix + '/lib/python%s/dist-packages/',            
            sys.prefix + '/local/lib/python%s/dist-packages/',            
            '/Library/Python/%s/site-packages/',
        ))

    for path in paths:
        if os.path.exists(path):
            return os.path.join(path, "quickstats", "macros")
    print('no installation path found', file=sys.stderr)
    return None

class CustomInstall(install):
    def run(self):
        install.run(self)
        try:
            import ROOT
            macros_dir = get_macros_dir()
            macros_path = os.path.join(macros_dir, "RooTwoSidedCBShape.cxx")
            if os.path.exists(macros_path):
                print('INFO: Found macros in "{}". Compiling...'.format(macros_path))
                ROOT.gROOT.LoadMacro('{}++'.format(macros_path))
                print('INFO: Successfully compiled ROOT macros.')
            else:
                print('WARNING: Unable to locate macro files from {}. ROOT macros will not be compiled.'.format(macros_dir))
        except:
            print('WARNING: PyROOT is not installed. ROOT macros will not be compiled.')


with open("README.md", "r") as fh:
    long_description = fh.read()
    
VERSIONFILE="quickstats/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))    

setuptools.setup(
    name="quickstats", # Replace with your own username
    version=verstr,
    author="Alkaid Cheng",
    author_email="chi.lung.cheng@cern.ch",
    description="A tool for quick statistical analysis for HEP experiments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={'quickstats':['macros/*.cxx', 'macros/*.h']},
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
          'numpy',
          'matplotlib',
          'click',
          'pandas',
          'uproot'
      ],
    scripts=['bin/quickstats'],
    python_requires='>=3.6',
)



from enum import Enum
from typing import Union
from functools import total_ordering

@total_ordering
class Verbosity(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        elif isinstance(other, int):
            return self.value < other
        elif isinstance(other, str):
            return self.value < getattr(self, other.upper()).value
        return NotImplemented

class VerbosePrint:
    @property
    def verbosity(self):
        return self._verbosity
    
    @verbosity.setter
    def verbosity(self, val):
        if isinstance(val, str):
            try:
                v = getattr(Verbosity, val.upper())
            except:
                raise ValueError("invalid verbosity level: {}".format(val))
            self._verbosity = v
        else:
            self._verbosity = val

    def __init__(self, verbosity:Union[int, Verbosity, str]=Verbosity.INFO):
        self.verbosity = verbosity
        
    def info(self, text:str):
        self.__call__(text, Verbosity.INFO)
        
    def warning(self, text:str):
        self.__call__(text, Verbosity.WARNING)
        
    def error(self, text:str):
        self.__call__(text, Verbosity.ERROR)
        
    def critical(self, text:str):
        self.__call__(text, Verbosity.CRITICAL)

    def debug(self, text:str):
        self.__call__(text, Verbosity.DEBUG)
        
    def __call__(self, text:str, verbosity:int=Verbosity.INFO):
        if verbosity >= self.verbosity:
            print(text)


import os
import json
import click

@click.command(name='run_pulls')
@click.option('-i', '--input_file', required=True, help='Path to the input workspace file')
@click.option('-w', '--workspace', default=None, help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', default=None, help='Name of model config. Auto-detect by default.')
@click.option('-d', '--data', default='combData', help='Name of dataset')
@click.option('-p', '--parameter', default='', help='Nuisance parameter(s) to run pulls on.'+\
                                                    'Multiple parameters are separated by commas.'+\
                                                    'Wildcards are accepted.')
@click.option('-x', '--poi', default="", help='POIs to measure')
@click.option('-r', '--profile', default="", help='Parameters to profile')
@click.option('-f', '--fix', default="", help='Parameters to fix')
@click.option('-s', '--snapshot', default="nominalNuis", help='Name of initial snapshot')
@click.option('-o', '--outdir', default="pulls", help='Output directory')
@click.option('-t', '--minimizer_type', default="Minuit2", help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, help='Number of CPUs to use per parameter')
@click.option('--binned/--unbinned', default=True, help='Binned likelihood')
@click.option('-q', '--precision', type=float, default=0.001, help='Precision for scan')
@click.option('-e', '--eps', type=float, default=1.0, help='Convergence criterium')
@click.option('-l', '--log_level', default="INFO", help='Log level')
@click.option('--eigen/--no-eigen', default=False, help='Compute eigenvalues and vectors')
@click.option('--print_level', type=int, default=1, help='Minimizer print level')
@click.option('--strategy', type=int, default=0, help='Default strategy')
@click.option('--fix-cache/--no-fix-cache', default=True, help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-multi', default=True, help='Fix MultiPdf level 2')
@click.option('--offset/--no-offset', default=True, help='Offset likelihood')
@click.option('--optimize', type=int, default=2, help='Optimize constant terms')
@click.option('--max_calls', type=int, default=-1, help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, help='Maximum number of Minuit iterations')
@click.option('--batch_mode/--no-batch', default=False, help='Batch mode when evaluating likelihood')
@click.option('--int_bin_precision', type=float, default=-1., help='Integrate the PDF over the bins '
                                                                   'instead of using the probability '
                                                                   'density at the bin centre')
@click.option('--parallel', type=int, default=0, help='Parallelize job across different nuisance'+\
                                                      'parameters using N workers.'+\
                                                      'Use -1 for N_CPU workers.')
@click.option('--cache/--no-cache', default=True, help='Cache existing result')
@click.option('--exclude', default="", help='Exclude NPs (wildcard is accepted)')
def run_pulls(**kwargs):
    """
    Tool for computing NP pulls and impacts
    """
    from quickstats.components import NuisanceParameterPull
    NuisanceParameterPull().run_pulls(**kwargs)
    
@click.command(name='plot_pulls')
@click.option('-i', '--inputdir', required=True, help='Path to directory containing pull results')
@click.option('-p', '--poi', default=None, help='Parameter of interest for plotting impact')
@click.option('-n', '--n_rank', type=int, default=None, help='Total number of NP to rank')
@click.option('-m', '--rank_per_plot', type=int, default=20, show_default=True,
              help='Number of NP to show in a single plot')
@click.option('--ranking/--no_ranking', default=True, show_default=True,
              help='Rank NP by impact')
@click.option('--threshold', type=float, default=0., show_default=True,
              help='Filter NP by postfit impact threshold')
@click.option('--show_sigma/--hide_sigma', default=True, show_default=True,
              help='Show one standard deviation pull')
@click.option('--show_prefit/--hide_prefit', default=True, show_default=True,
              help='Show prefit impact')
@click.option('--show_postfit/--hide_postfit', default=True, show_default=True,
              help='Show postfit impact')
@click.option('--sigma_bands/--no_sigma_bands', default=False, show_default=True,
              help='Draw +-1, +-2 sigma bands')
@click.option('--sigma_lines/--no_sigma_lines', default=True, show_default=True,
              help='Draw +-1 sigma lines')
@click.option('--shade/--no_shade', default=True, show_default=True,
              help='Draw shade')
@click.option('--correlation/--no_correlation', default=True, show_default=True,
              help='Show correlation impact')
@click.option('--onesided/--overlap', default=True, show_default=True,
              help='Show onesided impact')
@click.option('--relative/--absolute', default=False, show_default=True,
              help='Show relative variation')
@click.option('--theta_max', type=float, default=2, show_default=True,
              help='Pull range')
@click.option('-y', '--padding', type=int, default=7, show_default=True,
              help='Padding below plot for texts and legends. NP column height is 1 unit.')
@click.option('-h', '--height', type=float, default=1.0, show_default=True,
              help='NP column height')
@click.option('-s', '--spacing', type=float, default=0., show_default=True,
              help='Spacing between impact box')
@click.option('-d', '--display_poi', default=r"$\mu$", show_default=True,
              help='POI name to be shown in the plot')
@click.option('-t', '--extra_text', default=None, help='Extra texts below the ATLAS label. '+\
                                                       'Use "//" as newline delimiter')
@click.option('--elumi_label/--no_elumi_label', default=True, show_default=True,
              help='Show energy and luminosity labels')
@click.option('--ranking_label/--no_ranking_label', default=True, show_default=True,
              help='Show ranking label')
@click.option('--energy', type=float, default=13, show_default=True, help='Beam energy')
@click.option('--lumi', type=float, default=139, show_default=True, help='Luminosity')
@click.option('--combine_pdf/--split_pdf', default=True, show_default=True,
              help='Combine all ranking plots into a single pdf')
@click.option('--outdir', default='ranking_plots', show_default=True,
              help='Output directory')
@click.option('-o', '--outname', default='ranking', show_default=True,
              help='Output file name prefix')
@click.option('--style', default='default', show_default=True,
              help='Plotting style. Built-in styles are "default" and "trex".'+\
                   'Specify path to yaml file to set custom plotting style.')
@click.option('--fix_axis_scale/--free_axis_scale', default=True, show_default=True,
              help='Fix the axis scale across all ranking plots')
def plot_pulls(**kwargs):
    """
    Tool for plotting NP pulls and impact rankings
    """    
    from quickstats.plots.np_ranking_plot import NPRankingPlot
    inputdir, poi = kwargs.pop('inputdir'), kwargs.pop('poi')
    ranking_plot = NPRankingPlot(inputdir, poi)
    ranking_plot.plot(**kwargs)
    
    
@click.command(name='likelihood_scan')
@click.option('-i', '--input_file', required=True, help='Path to the input workspace file.')
@click.option('--min', 'poi_min', type=float, required=True, help='Minimum POI value to scan.')
@click.option('--max', 'poi_max', type=float, required=True, help='Maximum POI value to scan.')
@click.option('--step', 'poi_step', type=float, required=True, help='Scan interval.')
@click.option('-p', '--poi', default="", help='POI to scan. If not specified, the first POI from the workspace is used.')
@click.option('--cache/--no-cache', default=True, help='Cache existing result')
@click.option('-o', '--outname', default='{poi}', help='Name of output.')
@click.option('--outdir', default='likelihood_scan', help='Output directory.')
@click.option('--vmin', type=float, default=10, help='Minimum range of POI relative to the central value during likelihood calculation.')
@click.option('--vmax', type=float, default=10, help='Maximum range of POI relative to the central value during likelihood calculation.')
@click.option('-w', '--workspace', default=None, help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', default=None, help='Name of model config. Auto-detect by default.')
@click.option('-d', '--data', default='combData', help='Name of dataset.')
@click.option('-s', '--snapshot', default=None, help='Name of initial snapshot')
@click.option('-r', '--profile', default="", help='Parameters to profile')
@click.option('-f', '--fix', default="", help='Parameters to fix')
@click.option('--hesse/--no-hesse', default=False, help='Use Hesse error calculation')
@click.option('--minos/--no-minos', default=True, help='Use Minos error calculation')
@click.option('--constrain/--no-constrain', default=True, help='Use constrained NLL (i.e. include systematics)')
@click.option('-t', '--minimizer_type', default="Minuit2", help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, help='Number of CPUs to use per parameter')
@click.option('--binned/--unbinned', default=True, help='Binned likelihood')
@click.option('-e', '--eps', type=float, default=1.0, help='Convergence criterium')
@click.option('--strategy', type=int, default=0, help='Default minimization strategy')
@click.option('--print_level', type=int, default=1, help='Minimizer print level')
@click.option('--fix-cache/--no-fix-cache', default=True, help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, help='Fix MultiPdf level 2')
@click.option('--mpsplit',  default=3, help='MP split mode')
@click.option('-v', '--verbose',  default=0, help='verbosity')
@click.option('--max_calls', type=int, default=-1, help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, help='Maximum number of Minuit iterations')
@click.option('--optimize', type=int, default=2, help='Optimize constant terms')
@click.option('--offset/--no-offset', default=False, help='Offset likelihood')
@click.option('--parallel', type=int, default=-1, help='Parallelize job across different scan values.'+\
                                                       'Use -1 for N_CPU workers.')
def likelihood_scan(**kwargs):
    """
    Tool for likelihood scan
    """
    from quickstats.components.likelihood import scan_nll
    scan_nll(**kwargs)

@click.command(name='cls_limit')
@click.option('-i', '--input_file', 'filename', required=True, help='Path to the input workspace file')
@click.option('-p', '--poi', 'poi_name', default=None, help='POI to scan. If not specified, the first POI from the workspace is used.')
@click.option('-d', '--data', 'data_name', default='combData', help='Name of dataset')
@click.option('-o', '--outname', default='limits.json', help='Name of output')
@click.option('--blind/--unblind', 'do_blind', default=True, help='Blind/unblind analysis')
@click.option('--CL', 'CL', default=0.95, help='CL value to use')
@click.option('--precision', default=0.005, help='precision in mu that defines iterative cutoff')
@click.option('--do_tilde/--no_tilde', default=True, help='bound mu at zero if true and do the \tilde{q}_{mu} asymptotics')
@click.option('--predictive_fit/--no_predictive_fit', default=True, help='extrapolate best fit nuisance parameters based on previous fit results')
@click.option('--do_better_bands/--skip_better_bands', default=True, help='evaluate asymptotic CLs limit for various sigma bands')
@click.option('--better_negative_bands/--skip_better_negative_bands', default=False, 
              help='evaluate asymptotic CLs limit for negative sigma bands')
@click.option('--binned/--unbinned', 'binned_likelihood', default=True, help='Binned likelihood')
@click.option('--save_summary/--skip_summary', default=True, help='Save summary information')
@click.option('-f', '--fix', 'fix_param', default="", help='Parameters to fix')
@click.option('-r', '--profile', 'profile_param', default="", help='Parameters to profile')
@click.option('-w', '--workspace', 'ws_name', default=None, help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, help='Name of model config. Auto-detect by default.')
@click.option('-s', '--snapshot', 'snapshot_name', default=None, help='Name of initial snapshot')
@click.option('-t', '--minimizer_type', default="Minuit2", help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", help='Minimizer algorithm')
@click.option('-e', '--eps', type=float, default=1.0, help='Convergence criterium')
@click.option('--strategy', type=int, default=1, help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, help='Minimizer print level')
@click.option('--timer/--no_timer', default=False, help='Enable minimizer timer')
@click.option('-c', '--num_cpu', type=int, default=1, help='Number of CPUs to use per parameter')
@click.option('--offset/--no-offset', default=True, help='Offset likelihood')
@click.option('--optimize', type=int, default=2, help='Optimize constant terms')
@click.option('--fix-cache/--no-fix-cache', default=True, help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, help='Maximum number of Minuit iterations')
@click.option('--batch_mode/--no-batch', default=False, help='Batch mode when evaluating likelihood')
@click.option('--int_bin_precision', type=float, default=-1., help='Integrate the PDF over the bins '
                                                                   'instead of using the probability '
                                                                   'density at the bin centre')
@click.option('--constrain/--no-constrain', 'constrain_nuis', default=True, help='Use constrained NLL')
def cls_limit(**kwargs):
    """
    Tool for evaluating Asymptotic CLs limit
    """
    from quickstats.components import AsymptoticCLs
    outname = kwargs.pop('outname')
    save_summary = kwargs.pop('save_summary')
    asymptotic_cls = AsymptoticCLs(**kwargs)
    asymptotic_cls.evaluate_limits()
    asymptotic_cls.save(outname, summary=save_summary)

@click.command(name='compile')
@click.option('-m', '--macros', default=None, help='Macros to compile (separated by commas)')
def compile_macros(macros):
    """
    Compile ROOT macros
    """
    from quickstats.utils.root_utils import compile_macro
    if macros is None:
        macros = ["RooTwoSidedCBShape", "AsymptoticCLsTool", "RooFitObjects"]
    else:
        macros = macros.split(',')
    for macro in macros:
        compile_macro(macro)
    
@click.command(name='harmonize_np')
@click.argument('ws_files', nargs=-1)
@click.option('-r', '--reference', required=True, help='Path to reference json file containing renaming scheme')
@click.option('-i', '--input_config_path', default=None, help='Path to json file containing input workspace paths')
@click.option('-b', '--base_path', default='./', help='Base path for input config')
@click.option('-o', '--outfile', default='renamed_np.json', help='Output filename')
def harmonize_np(ws_files, reference, input_config_path, base_path, outfile):
    """
    Harmonize NP names across different workspaces
    """
    from quickstats.components import NuisanceParameterHarmonizer
    harmonizer = NuisanceParameterHarmonizer(reference)
    if (len(ws_files) > 0) and input_config_path is not None:
        raise RuntimeError('either workspace paths or json file containing workspace paths should be given')
    if len(ws_files) > 0:
        harmonizer.harmonize(ws_files, outfile=outfile)
    elif (input_config_path is not None):
        harmonizer.harmonize_multi_input(input_config_path, base_path, outfile=outfile)
        
        
@click.command(name='generate_asimov')
@click.option('-i', '--input_file', required=True, help='Path to the input workspace file.')
@click.option('-o', '--output_file', required=True, help='Name of the output workspace containing the '
                                                         'generated asimov dataset.')
@click.option('-p', '--poi', required=True, help='Name of the parameter of interest (POI).')
@click.option('--poi_val', type=float, required=True,
              help='Generate asimov data with POI set at the specified value.')
@click.option('--poi_profile', type=float, default=None,
              help='Perform NP profiling with POI set at the specified value.')
@click.option('--globs_np_matching/--no_globs_np_matching', default=True, show_default=True,
              help='Match the values of nuisance parameters and the corresponding global '
                   'observables when generating the asimov data. This is important for making '
                   'sure the asimov data has the (conditional) minimal NLL.')
@click.option('--conditional_mle/--unconditional_mle', default=True, show_default=True,
              help='Perform conditional maximum likelihood estimation. This option is effective '
                   'only if poi_profile is specified. If True, best-fit values are obtained for '
                   'mu = poi_profile. If False, the overal best-fit values are obtained.')
@click.option('--asimov_name', default="asimovData_{mu}", show_default=True,
              help='Name of the generated asimov dataset.')
@click.option('-d', '--data', default='combData', show_default=True,
              help='Name of the dataset used in NP profiling.')
@click.option('-c', '--configuration', default=None,
              help='Path to the json configuration file containing'
                   ' the minimizer options for NP profiling.')
def generate_asimov(**kwargs):
    """
    Generate Asimov dataset
    """
    input_filename = kwargs.pop("input_file")
    output_filename = kwargs.pop("output_file")
    if os.path.abspath(input_filename) == os.path.abspath(output_filename):
        raise ValueError("output workspace file name cannnot be the same as the input workspace file name")
    from quickstats.components import AnalysisObject
    config_file = kwargs.pop("configuration")
    if config_file is not None:
        config = json.load(open(config_file, 'r'))
    else:
        config = {}
    data_name = kwargs.pop("data")
    kwargs['poi_name'] = kwargs.pop("poi")
    task = AnalysisObject(filename=input_filename, data_name=data_name, **config)
    task.model.generate_asimov(**kwargs)
    task.model.save(output_filename)


##################################################################################################
# Based on https://gitlab.cern.ch/atlas-physics/stat/tools/StatisticsTools by Stefan Gadatsch
# Author: Alkaid Cheng
# Email: chi.lung.cheng@cern.ch
##################################################################################################
from typing import Union, Optional, List, Dict
import math

import numpy as np
import ROOT

class ExtendedMinimizer(object):
    _SCAN_EXCLUDE_CONFIG_ = ['scan', 'minos', 'save']
    _NLL_COMMANDS_ = ['NumCPU', 'Constrain', 'CloneData', 'GlobalObservables', 
                      'GlobalObservablesTag', 'OffsetLikelihood', 'Offset']

    _DEFAULT_MINIMIZER_OPTION_ = {
        'minimizer_type': 'Minuit2',
        'minimizer_algo': 'Migrad',        
        'opt_const': 2,
        'verbose': 0,
        'save': 0,
        'timer': 0,
        'hesse': 0,
        'minos': 0,
        'scan': 0,
        'num_ee': 5,
        'do_ee_wall': 1,
        'retry': 0,
        'eigen': 0,
        'reuse_minimizer': 0,
        'reuse_nll': 0,
        'eps': 1.0,
        'max_calls': -1,
        'max_iters': -1,
        'n_sigma': 1,
        'precision': 1,
        'default_strategy': 1,
        'print_level': -1
    }
    
    _DEFAULT_NLL_OPTION_ = {
        'num_cpu': 1,
        'offset': True
    }
    
    @property
    def config(self):
        return self._config
    
    def __init__(self, minimizer_name:str, pdf:"ROOT.RooAbsPdf", data:"ROOT.RooAbsData",
                 minimizer_options:Optional[Dict]=None):
        """
            Minimizr Options:
                precision: precision for the find sigma algorithm
        """
        self.name = minimizer_name
        self.pdf = pdf
        self.data = data
        self.minimizer = None
        self.nll = None
        self.fit_result = None
        
        self._config = self._DEFAULT_MINIMIZER_OPTION_ 
        if minimizer_options is not None:
            self._config.update(minimizer_options)
        self.minos_set = ROOT.RooArgSet()
        self.cond_set  = ROOT.RooArgSet()
        self.scan_set  = ROOT.RooArgSet()         
        
        self.Hessian_matrix = None
        self.min_nll = None
        
        self.eigen_values = ROOT.TVectorD()
        self.eigen_vectors = ROOT.TMatrixD()
        
        self.fit_options = {}
        self.scan_options = {}
        self.nll_command_list = ROOT.RooLinkedList()
        
        self.configure_default_minimizer_options()
        
        print('INFO: Created ExtendedMinimizer("{}") instance'.format(self.name))
        
    @property
    def data(self):
        return self._data
    
    @data.setter
    def data(self, val):
        assert isinstance(val, ROOT.RooAbsData)
        self._data = val
    
    @property
    def pdf(self):
        return self._pdf

    @pdf.setter
    def pdf(self, val):
        assert isinstance(val, ROOT.RooAbsPdf)
        self._pdf = val 
            
    @staticmethod
    def _configure_default_minimizer_options(minimizer_type='Minuit2', minimizer_algo='Migrad',
                                             default_strategy=0, print_level=-1,
                                             debug_mode=False):
        ROOT.Math.MinimizerOptions.SetDefaultMinimizer(minimizer_type, minimizer_algo)
        ROOT.Math.MinimizerOptions.SetDefaultStrategy(default_strategy)
        ROOT.Math.MinimizerOptions.SetDefaultPrintLevel(print_level)
        if debug_mode:
            ROOT.Math.MinimizerOptions.SetDefaultPrintLevel(1)
        else:
            ROOT.Math.MinimizerOptions.SetDefaultPrintLevel(-1)
            ROOT.RooMsgService.instance().setGlobalKillBelow(ROOT.RooFit.FATAL)
        if ROOT.Math.MinimizerOptions.DefaultPrintLevel() < 0:
            ROOT.RooMsgService.instance().setGlobalKillBelow(ROOT.RooFit.FATAL)
        return None
            
    def configure_default_minimizer_options(self):
        self._configure_default_minimizer_options(self.config['minimizer_type'],
                                                  self.config['minimizer_algo'],
                                                  self.config['default_strategy'],
                                                  self.config['print_level'])
        return None
    
    def create_nll(self, nll_command_list:List["ROOT.RooCmdArg"]=None):
        if (not self.nll) or (not self.config['reuse_nll']):
            if nll_command_list is None:
                nll_command_list = self.nll_command_list
            else:
                nll_command_list = self.parse_nll_commands(nll_command_list)
            #self.nll = self.pdf.createNLL(self.data, nll_command_list)
            self.nll = self.pdf.createNLL(self.data, *nll_command_list)
        
    def create_minimizer(self):
        
        if not self.nll:
            raise RuntimeError('NLL not initialized')

        if (self.minimizer is None) or (not self.config['reuse_minimizer']):
            self.configure_default_minimizer_options()
            self.minimizer = ROOT.RooMinimizer(self.nll)
            self.minimizer.setPrintLevel(self.config['print_level'])
            self.minimizer.optimizeConst(self.config['opt_const'])
            self.minimizer.setMinimizerType(self.config['minimizer_type'])
            self.minimizer.setEvalErrorWall(self.config['do_ee_wall'])
            self.minimizer.setPrintEvalErrors(self.config['num_ee'])
            self.minimizer.setVerbose(self.config['verbose'])
            self.minimizer.setProfile(self.config['timer'])
            self.minimizer.setStrategy(self.config['default_strategy'])
            self.minimizer.setEps(self.config['eps'])
            
            if self.config['max_calls'] != -1:
                self.minimizer.setMaxFunctionCalls(self.config['max_calls'])
            if self.config['max_iters'] != -1:
                self.minimizer.setMaxIterations(self.config['max_iters'])
 
    def parse_nll_commands(self, commands:List["ROOT.RooCmdArg"]):
        # temporary fix for ROOT version below 6.22/06
        #nll_command_list = ROOT.RooLinkedList()
        #for command in commands:
        #    nll_command_list.Add(command)
        #return nll_command_list
        return commands
        
    def configure_nll(self, constrain=None, global_observables=None, **kwargs):
        commands = []
        if 'num_cpu' in kwargs:
            commands.append(ROOT.RooFit.NumCPU(kwargs['num_cpu'], 3))
        if 'offset' in kwargs:
            commands.append(ROOT.RooFit.Offset(kwargs['offset']))
        if constrain is not None:
            commands.append(ROOT.RooFit.Constrain(constrain))
        if global_observables is not None:
            commands.append(ROOT.RooFit.GlobalObservables(global_observables))            
        self.nll_command_list = self.parse_nll_commands(commands)
    
    def configure(self, **kwargs):
        minos_set = kwargs.pop('minos_set', None)
        cond_set  = kwargs.pop('cond_set' , None)
        scan_set  = kwargs.pop('scan_set' , None)
        if minos_set is not None:
            self.minos_set = minos_set
        if cond_set is not None:
            self.cond_set = cond_set
        if scan_set is not None:
            self.scan_set = scan_set
        # parse nll commands 
        nll_commands = kwargs.pop('nll_commands', [])
        if nll_commands:
            self.nll_command_list = self.parse_nll_commands(nll_commands)
        
        # configure minimizer settings
        for arg in kwargs:
            if arg not in self._config:
                raise ValueError('{} is not a valid minimizer config'.format(arg))
            self._config[arg] = kwargs[arg]
        
        self.fit_options = kwargs
        self.scan_options = {k:v for k,v in kwargs.items() if k not in self._SCAN_EXCLUDE_CONFIG_}

    def minimize(self, nll=None, **kwargs):

        self.configure(**kwargs)
        if nll is None:
            self.create_nll()
        else:
            self.nll = nll
        
        if self.cond_set.size() > 0:
            attached_set = self.nll.getVariables()
            for cond in self.cond_set:
                buffer = attached_set.find(cond.GetName())
                if buffer:
                    buffer.setVal(cond.getVal())
                    buffer.setConstant(1)
                    
        self.create_minimizer()
        status = 0
        attached_set = self.nll.getVariables()
        perform_minimization = any(not attached.isConstant() for attached in attached_set)

        if not perform_minimization:
            print('INFO: ExtendedMinimizer::minimize("{}") no floating parameters found'
                  '-- skipping minimization'.format(self.name))
        else:
            status = self.robust_minimize()
        
        # Evaluate errors with Hesse
        if self.config['hesse']:
            self.minimizer.hesse()
        
        # Obtain Hessian matrix either from patched Minuit or after inversion
        # TMatrixDSym G = Minuit2::MnHesse::lastHessian();
        last_fit = self.minimizer.lastMinuitFit()
        if last_fit:
            self.Hessian_matrix = last_fit.covarianceMatrix().Invert()
        else:
            self.Hessian_matrix = None
        
        # Eigenvalue and eigenvector analysis
        if self.config['eigen']:
            self.eigen_analysis()
        
        # Evaluate errors with Minos
        if self.config['minos']:
            if self.minos_set.size() > 0:
                self.minimizer.minos(self.minos_set)
            else:
                self.minimizer.minos()
        
        self.min_nll = self.nll.getVal()
        
        if self.config['scan']:
            self.find_sigma()

        if self.config['save']:
            data_name = self.data.GetName()
            save_name = "fitresult_{}_{}".format(self.name, data_name)
            save_title = "Result of fit of p.d.f. {} to dataset {}".format(self.name, data_name)
            print('INFO: ExtendedMinimizer::minimize("{}") saving results as {}'.format(self.name, save_name))
            self.fit_result = self.minimizer.save(save_name, save_title)
        # if the 'scan' option is used, minimizer will get deleted before this line
        elif self.minimizer is not None:
            self.fit_result = self.minimizer.save()

        if self.cond_set.size() > 0:
            attached_set = self.nll.getVariables()
            for cond in self.cond_set:
                buffer = attached_set.find(cond.GetName())
                if buffer:
                    buffer.setVal(cond.getVal())
                    buffer.setConstant(cond.isConstant())

        if not self.config['reuse_minimizer']:
            self.minimizer = None
        
        return status
    
    def eigen_analysis(self):
        if not isinstance(self.Hessian_matrix, ROOT.TMatrixDSym):
            raise ValueError('invalid Hessian matrix')
        n = self.Hessian_matrix.GetNrows()
        
        # construct reduced Hessian matrix
        Gred = ROOT.TMatrixDSym(n)
        for i in range(n):
            for j in range(n):
                norm = math.sqrt(self.Hessian_matrix(i, i)*self.Hessian_matrix(j, j))
                Gred[i][j] = self.Hessian_matrix(i, j)/norm
        
        # perform eigenvalue analysis using ROOT standard tools
        Geigen = ROOT.TMatrixDSymEigen(Gred)
        
        self.eigen_values = Geigen.GetEigenValues()
        self.eigen_vectors = Geigen.GetEigenVectors()
        
        # simple printing of eigenvalues and eigenvectors
        self.eigen_values.Print()
        self.eigen_vectors.Print()
    
    def robust_minimize(self):

        strategy = self.config['default_strategy']
        retry = self.config['retry']
        minimizer_type = self.config['minimizer_type']
        minimizer_algo = self.config['minimizer_algo']

        status = self.minimizer.minimize(minimizer_type, minimizer_algo)
        
        while ((status != 0) and (status!=1) and (strategy<2) and (retry>0)):
            strategy += 1
            retry -= 1
            print('ExtendedMinimizer::robust_minimize("{}") fit failed with status {}. '
                  'Retrying with strategy {}'.format(self.name, status, strategy))
            self.minimizer.setStrategy(strat)
            status = self.minimizer.minimize(minimizer_type, minimizer_algo)
        
        if status not in [0, 1]:
            print('ExtendedMinimizer::robust_minimize("{}") fit failed with status {}'.format(self.name, status))
        
        self.minimizer.setStrategy(self.config['default_strategy'])
        
        return status
    
    def use_limits(self, par:"ROOT.RooRealVar", val:float):
        if (val < par.getMin()):
            print('ExtendedMinimizer::use_limits("{}") {} = {} limited by minimum at {}'.format(
                   self.name, par.GetName(), val, par.getMin()))
            return par.getMin()
        elif (val > par.getMax()):
            print('ExtendedMinimizer::use_limits("{}") {} = {} limited by maximum at {}'.format(
                   self.name, par.GetName(), val, par.getMax()))
            return par.getMax()
        else:
            return val
    
    def find_sigma(self):
        if self.scan_set.getSize() == 0:
            attached_set = self.nll.getVariables()
            ROOT.RooStats.RemoveConstantParameters(attached_set)
            self.scan_set = attached_set

        for v in self.scan_set:
            variables = self.pdf.getVariables()
            ROOT.RooStats.RemoveConstantParameters(variables)
            variables.add(v, ROOT.kTRUE)
            snapshot = variables.snapshot()
            
            self.config['n_sigma'] = abs(self.config['n_sigma'])
            val = v.getVal()
            err = self.config['n_sigma'] * v.getError()
            
            min_nll = self.min_nll
            hi = self._find_sigma(min_nll, val+err, val, v, 
                                  self.scan_options, self.config['n_sigma'],
                                  self.config['precision'], self.config['eps'])
            variables.__assign__(snapshot)
            lo = self._find_sigma(min_nll, val-err, val, v, 
                                  self.scan_options, -self.config['n_sigma'],
                                  self.config['precision'], self.config['eps'])
            variables.__assign__(snapshot)
            self.min_nll = min_nll
            
            _lo = lo if not math.isnan(lo) else 1.0
            _hi = hi if not math.isnan(hi) else -1.0
            v.setAsymError(_lo, _hi)
            
            print('ExtendedMinimizer::minimize("{}")'.format(self.name))        
            print(v)

    # _____________________________________________________________________________
    # Find the value of sigma evaluated at a specified nsigma, assuming NLL -2logL is roughly parabolic in par.
    # The precision is specified as a fraction of the error, or based on the Minuit default tolerance.
    # Based on https://svnweb.cern.ch/trac/atlasoff/browser/PhysicsAnalysis/HiggsPhys/HSG3/WWDileptonAnalysisCode/HWWStatisticsCode/trunk/macros/findSigma.C
    # by Aaron Armbruster <aaron.james.armbruster@cern.ch> and adopted by Tim Adye <T.J.Adye@rl.ac.uk>.
    def _find_sigma(self, nll_min:float, val_guess:float, val_mle:float,
                    par:"ROOT.RooRealVar", scan_options:Dict, n_sigma:float,
                    precision:float, fit_tol:float):
        '''
        args:
            precision: fit precision
            fit_tol: fit tolerance
        '''
        max_iter = 25

        param_name = par.GetName()
        val_guess = self.use_limits(par, val_guess)
        direction = +1 if n_sigma >= 0.0 else -1
        n_damping = 1
        damping_factor = 1.0
        guess_to_corr = {}
        t_mu = ROOT.TMath.QuietNaN()
        front_str = 'ExtendedMinimizer::findSigma("{}") '.format(self.name)
        
        
        if precision <= 0.0:
            # RooFit default tolerance is 1.0
            tol = fit_tol if fit_tol > 0.0 else 1.0
            eps = 0.001 * tol
            precision = 5.0*eps / (n_sigma**2)
        
        temp_options = {**scan_options}
        snapshot = self.cond_set.snapshot()
        for i in range(max_iter):
            print('{} Parameter {} {}sigma '
                  'iteration {}: start {} (MLE{})'.format(front_str, par.GetName(),
                    n_sigma, i+1, val_guess, val_guess-val_mle))
            val_pre = val_guess
            
            poi_set = ROOT.RooArgSet(par).snapshot()
            poi_set.first().setVal(val_guess)
            
            scan_options['reuse_nll'] = 1
            scan_options['scan']      = 0
    
            self.minimize(cond_set=poi_set, 
                          **scan_options)
            
            self.cond_set = snapshot
            scan_options.clear()
            scan_options.update(temp_options)
            
            nll = self.min_nll
            poi_set = None
            
            t_mu = 2.0 * (nll - nll_min)
            sigma_guess = abs(val_guess-val_mle)
            if (t_mu > 0.01):
                sigma_guess /= math.sqrt(t_mu)
            else:
                sigma_guess *= 10.0 # protect against t_mu <=0 and also don't move too far
            
            corr = damping_factor*(val_pre - val_mle - n_sigma*sigma_guess)
            
            for guess in guess_to_corr:
                if (abs(guess - val_pre) < direction*val_pre*0.02):
                    damping_factor *= 0.8
                    print('{} Changing damping factor to {}'.format(front_str, damping_factor))
                    n_damping += 1
                    if n_damping > 10:
                        n_damping = 1
                        damping_factor = 1.0
                    corr *= damping_factor
                    break
            
            # subtract off the difference in the new and damped correction
            val_guess -= corr
            guess_to_corr[val_pre] = corr          
            val_guess = self.use_limits(par, val_guess)
            rel_precision = precision*abs(val_guess-val_mle)
            delta = val_guess - val_pre
            
            
            print('{} {} {:.3f} (MLE {:.3f}) -> {:.3f} (MLE {:.3f}), '
                  'change {:.3f}, precision {:.3f}, -2lnL {:.4f}, sigma(guess) {:.3f})'.format(
                  front_str, param_name, val_pre, val_pre - val_mle, val_guess, val_guess - val_mle,
                  delta, rel_precision, t_mu, sigma_guess))
            print('{} NLL:                 {}'.format(front_str, nll))
            print('{} delta(NLL):          {}'.format(front_str, nll - nll_min))
            print('{} nsigma*sigma(pre):   {}'.format(front_str, abs(val_pre - val_mle)))
            print('{} sigma(guess):        {}'.format(front_str, sigma_guess))
            print('{} par(guess):          {}'.format(front_str, val_guess + corr))
            print('{} best-fit val:        {}'.format(front_str, val_mle))
            print('{} tmu:                 {}'.format(front_str, t_mu))
            print('{} Precision:           {}'.format(front_str, direction*val_guess*precision))
            print('{} Correction:          {}'.format(front_str, -corr))
            print('{} nsigma*sigma(guess): {}'.format(front_str, abs(val_guess-val_mle)))
            
            if abs(delta) <= rel_precision:
                break
        if i >= max_iter:
            print('ERROR: find_sigma failed after {} iterations'.format(i+1))
            return ROOT.TMath.QuietNan()
        
        err = val_guess - val_mle
        print('{} {} {}sigma = {:.3F} at -2lnL = {:4f} after {} iterations'.format(
              front_str, par.GetName(), n_sigma, err, t_mu, i+1))
        
        return err
    
    def create_profile(self, var:"ROOT.RooRealVar", lo:float, hi:float, n_bins:int):
        map_poi2nll = {}
        
        ROOT.Math.MinimizerOptions.SetDefaultPrintLevel(-1)
        ROOT.RooMsgService.instance().setGlobalKillBelow(ROOT.RooFit.FATAL)
        
        variables = self.pdf.getVariables()
        ROOT.RooStats.RemoveConstantParameters(variables)
        
        # unconditional fit
        buffer_scan_options = deepcopy(scan_options)
        
        buffer_scan_options['reuse_nll'] = 1
        
        self.minimize(**buffer_scan_options)
        
        map_poi2nll[var.getVal()] = 2 * self.min_nll
        snapshot = variables.snapshot()
        
        ## these lines should be improved
        buffer_scan_options.pop('reuse_nll')
        scan_options.clear()
        scan_options.update(buffer_scan_options)
        
        # perform the scan
        delta_x = (hi - lo)/n_bins
        for i in range(n_bins):
            variables.__assign__(snapshot)
            var.setVal(lo + i * delta_x)
            val.setConstant(1)
            
            buffer_scan_options['reuse_nll'] = 1
            
            self.minimize(**buffer_scan_options)
            map_poi2nll[var.getVal()] = 2 * self.min_nll
            
            ## these lines should be improved
            buffer_scan_options.pop('reuse_nll')
            scan_options.clear()
            scan_options.update(buffer_scan_options)
            
            var.setConstant(0)
        variables.__assign__(snapshot)
        graphs = self.prepare_profile(map_poi2nll)
        return graphs
    
    def prepare_profile(self, map_poi2nll:Dict):
        x, y = [], []
        n_bins = len(map_poi2nll) - 1
        
        xlo = float('inf') 
        xhi = float('-inf') 
        
        for nll_prev in map_poi2nll:
            nll = map_poi2nll[nll_prev]
            if (not math.isinf(nll)) and (abs(nll) < 10**20):
                x.append(nll_prev)
                y.append(nll)
        
        nr_points = len(x)
        
        if nr_points == 0:
            raise ValueError("map_poi2nll is empty")
            
        x, y = zip(*sorted(zip(x, y)))
        x = np.array(x, dtype=float)
        y = np.array(y, dtype=float)
            
        if (x[0] < xlo):
            xlo = x[0]
            
        if x[-1] > xhi:
            xhi = x[-1]
        
        g = ROOT.TGraph(nr_points, x, y)
        
        min_nll = ROOT.TMath.Infinity()
        min_nll_x = 0.0
        
        for i in range(g.GetN()):
            tmpX, tmpY = ROOT.Double(0), ROOT.Double(0)
            g.GetPoint(i, tmpX, tmpY)
            if (tmpY < min_nll):
                min_nll = tmpY
                min_nll_x = tmpX
                
        for i in range(g.GetN()):
            tmpX, tmpY = ROOT.Double(0), ROOT.Double(0)
            g.GetPoint(i, tmpX, tmpY)
            tmpY -= min_nll
            g.SetPoint(i, tmpX, tmpY)

        min_nll = ROOT.TMath.Infinity()
        min_nll_x = 0.0
        
        # Make smooth interpolated graph for every folder in poi range, find minimum nll
        x_interpolated_coarse, y_interpolated_coarse = [], []
        
        stepsize_coarse = abs(xhi - xlo) / n_bins
        tmpX = xlo
        while (tmpX <= xhi):
            tmpY = g.Eval(tmpX, 0)
            x_interpolated_coarse.append(tmpX)
            y_interpolated_coarse.append(tmpY)
            tempX += stepsize_coarse
            
        x_interpolated_coarse = np.array(x_interpolated_coarse, dtype=float)
        y_interpolated_coarse = np.array(y_interpolated_coarse, dtype=float)
        nr_points_interpolated_coarse = len(x_interpolated_coarse)
        g_interpolated_coarse = ROOT.TGraph(nr_points_interpolated_coarse,
                                            x_interpolated_coarse,
                                            y_interpolated_coarse)
        
        x_interpolated, y_interpolated = [], []
        two_step_interpolation = False
        stepsize = abs(xhi - xlo) / (10*n_bins)
        while (tmpX <= xhi):
            tmpY = 0.0
            if two_step_interpolation:
                tmpY = g_interpolated_coarse.Eval(tmpX, 0, "S")
            else:
                tmpY = g.Eval(tmpX, 0, "S")
            x_interpolated.append(tmpX)
            y_interpolated.append(tmpY)
            tempX += stepsize
            
        x_interpolated = np.array(x_interpolated, dtype=float)
        y_interpolated = np.array(y_interpolated, dtype=float)
        nr_points_interpolated = len(x_interpolated)
        g_interpolated = ROOT.TGraph(nr_points_interpolated,
                                     x_interpolated,
                                     y_interpolated)
        
        for i in range(g_interpolated.GetN()):
            tmpX, tmpY = ROOT.Double(0), ROOT.Double(0)
            g_interpolated.GetPoint(i, tmpX, tmpY)
            if (tmpY < min_nll):
                min_nll = tmpY
                min_nll_x = tmpX
        
        for i in range(g.GetN()):
            tmpX, tmpY = ROOT.Double(0), ROOT.Double(0)
            g.GetPoint(i, tmpX, tmpY)
            tmpY -= min_nll
            g.SetPoint(i, tmpX, tmpY)
        
        for i in range(g_interpolated.GetN()):
            tmpX, tmpY = ROOT.Double(0), ROOT.Double(0)
            g_interpolated.GetPoint(i, tmpX, tmpY)
            tmpY -= min_nll
            g_interpolated.SetPoint(i, tmpX, tmpY)
        
        g.SetLineWidth(2)
        g.SetMarkerStyle(20)
        
        g_interpolated.SetLineWidth(2)
        g_interpolated.SetMarkerStyle(20)
        
        return (g, g_interpolated)
